
Require Import Arith.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import StaticWeak.
Require Import Substitution.
Require Import Weakening.
Require Import Compose.
Require Import Expand.
Require Import Size.
Require Import Similar.



Lemma can_compose :
  forall G A B s1 s2,
    wsof G s1 A (susp B)
    -> exists s, compose s1 s2 s.
Proof.
intros G A B s1 s2 Hof.
remember (susp B) as U eqn:HeqU.
revert B s2 HeqU.
induct Hof.

(* nl *)
intros G A B s2 HeqU.
exists s2; apply compose_nl; done.

(* pm *)
intros G p r A U G' _ _ B s2 HeqU.
discriminate HeqU.

(* app *)
intros G v s1 p A B U _ _ IH C s2 ->.
so (IH _ s2 (eq_refl _)) as (s & Hcompose).
exists (app v s); eauto using @compose; done.

(* pi1 *)
intros G s1 A B U _ IH C s2 ->.
so (IH _ s2 (eq_refl _)) as (s & Hcompose).
exists (pi1 s); eauto using @compose; done.

(* pi2 *)
intros G s1 A B U _ IH C s2 ->.
so (IH _ s2 (eq_refl _)) as (s & Hcompose).
exists (pi2 s); eauto using @compose; done.
Qed.


Lemma can_rcompose :
  forall G A r s,
    wrof G r (susp A)
    -> exists r', compose r s r'.
Proof.
intros G A r s2 Hof.
invertc Hof; [].
intros h s1 B B' _ _ Hofs1 <-.
so (can_compose _#4 s2 Hofs1) as (s & Hcompose).
exists (use h s); eauto using @compose; done.
Qed.

       
Lemma compose_formation_weak :
  forall G A B U s1 s2 s,
    compose s1 s2 s
    -> wsof G s1 A (susp B)
    -> wsof G s2 B U
    -> wsof G s A U.
Proof.
cut (forall z (h : z = sp) G A B U (s1 : exp z) s2 s,
       compose s1 s2 s
       -> wsof G (eq_rect _ _ s1 _ h) A (susp B)
       -> wsof G s2 B U
       -> wsof G (eq_rect _ _ s _ h) A U).
  intro Hind.
  intros; eapply (Hind sp (eq_refl _)); eauto; done.
intros z h G A B U s1 s2 s Hcompose Hofs1 Hofs2.     
revert h G A B U Hofs1 Hofs2.
#5 induct Hcompose; revertnew (intros; substrefls; try discriminate).

(* nl *)
intros s G A B U Hofs2 Hofs1.
invert Hofs1; [].
intros <-.
exact Hofs2.

(* pm *)
intros p r1 s2 r _ _ G A B U Hofs2 Hofs1.
invert Hofs1; done.

(* app *)
intros v s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros p A B Hofv Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
apply wsof_app; auto; done.

(* pi1 *)
intros s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros A B Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
apply wsof_pi1; auto; done.

(* pi2 *)
intros s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros A B Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
apply wsof_pi2; auto; done.
Qed.


Lemma rcompose_formation_weak :
  forall G A U r s r',
    compose r s r'
    -> wrof G r (susp A)
    -> wsof G s A U
    -> wrof G r' U.
Proof.
intros G A U r s2 r' Hcompose Hofr Hofs2.
invertc Hofr; [].
intros h s1 B B' Hofh Hsim Hofs1 <-.
invert Hcompose; [].
intros s Hcompose' <-.
so (compose_formation_weak _#7 Hcompose' Hofs1 Hofs2) as Hofs.
eapply wrof_use; eauto; done.
Qed.


Lemma can_expand :
  (forall A r, nres r -> exists m, expandm A r m)
  /\
  (forall A, exists p v, expandp A p v).
Proof.
cut (forall z (A : exp z),
       (forall (h : z = nt) r, nres r -> exists m, expandm (eq_rect _ _ A _ h) r m)
       /\
       (forall (h : z = pt), exists p v, expandp (eq_rect _ _ A _ h) p v)).
  intros Hind.
  split.
    intro A.
    exact (Hind _ A andel (eq_refl _)).

    intro A.
    exact (Hind _ A ander (eq_refl _)).
intros z A.
#9 induct A; revertnew (intros; split; intros; try discriminate; substrefls).

(* atom *)
intros t e IH r His.
exists (etam r); eauto using expandm.

(* up *)
intros e IH r His.
so (IH ander (eq_refl _)) as (p & v & Hexpand).
simpl in Hexpand.
so (can_compose_nres _ (pm p (ret v)) His) as (r' & Hcompose).
exists (shiftp r'); eauto using expandm; done.

(* pi *)
intros p A IH1 B IH2 r His.
so (IH1 ander (eq_refl _)) as (q & v & HexpandA).
simpl in HexpandA.
so (can_compose_nres _ (app v nl) (sh_nres 0 (bound q) _ His)) as (r' & Hcompose).
so (compose_nres_nspine _#3 Hcompose (sh_nres _#3 His) (nspine_app nspine_nl)) as Hisr'.
so (IH2 andel (eq_refl _) _ Hisr') as (m & HexpandB).
simpl in HexpandB.
exists (lam q m); eauto using expandm; done.

(* wth *)
intros A IH1 B IH2 r His.
so (can_compose_nres _ (pi1 nl) His) as (r1 & Hcompose1).
so (can_compose_nres _ (pi2 nl) His) as (r2 & Hcompose2).
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi1 nspine_nl)) as Hisr1.
so (compose_nres_nspine _#3 Hcompose2 His (nspine_pi2 nspine_nl)) as Hisr2.
so (IH1 andel (eq_refl _) r1 Hisr1) as (m1 & Hexpand1).
so (IH2 andel (eq_refl _) r2 Hisr2) as (m2 & Hexpand2).
simpl in Hexpand1, Hexpand2.
exists (pairm m1 m2); eauto using expandm; done.

(* sigma (same as wth) *)
intros A IH1 B IH2 r His.
so (can_compose_nres _ (pi1 nl) His) as (r1 & Hcompose1).
so (can_compose_nres _ (pi2 nl) His) as (r2 & Hcompose2).
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi1 nspine_nl)) as Hisr1.
so (compose_nres_nspine _#3 Hcompose2 His (nspine_pi2 nspine_nl)) as Hisr2.
so (IH1 andel (eq_refl _) r1 Hisr1) as (m1 & Hexpand1).
so (IH2 andel (eq_refl _) r2 Hisr2) as (m2 & Hexpand2).
simpl in Hexpand1, Hexpand2.
exists (pairm m1 m2); eauto using expandm; done.

(* down *)
intros A IH.
so (IH andel (eq_refl _) (use (var 0) nl) (nres_use nspine_nl)) as (m & Hexpand).
simpl in Hexpand.
exists shiftm, (thunk m); eauto using expandp; done.

(* one *)
exists unitpat, unitp; eauto using expandp; done.

(* tens *)
intros A IH1 B IH2.
so (IH1 ander (eq_refl _)) as (p & v & HexpandA).
so (IH2 ander (eq_refl _)) as (q & w & HexpandB).
simpl in HexpandA, HexpandB.
exists (pairpat p q), (pairp (sh 0 (bound q) v) w); eauto using expandp; done.

(* ex (almost same as tens) *)
intros p' A IH1 B IH2.
so (IH1 ander (eq_refl _)) as (p & v & HexpandA).
so (IH2 ander (eq_refl _)) as (q & w & HexpandB).
simpl in HexpandA, HexpandB.
exists (pairpat p q), (pairp (sh 0 (bound q) v) w); eauto using expandp; done.
Qed.


Lemma can_expandm :
  forall A r,
    nres r
    -> exists m, expandm A r m.
Proof.
exact (can_expand andel).
Qed.


Lemma can_expandp :
  forall A,
    exists p v, expandp A p v.
Proof.
exact (can_expand ander).
Qed.


Lemma can_pof_expandp_similar :
  forall A p v,
    expandp A p v
    -> exists A' G,
         similar A A'
         /\ pof p A' G.
Proof.
intros A p v Hexpand.
induct Hexpand; intros; destruct_everywhere; eexists; eexists; eauto using pof with similar.
Qed.


Lemma identity_expansion_weak :
  (forall G r A m,
     wrof G r (susp A)
     -> expandm A r m
     -> wtof G m A)
  /\
  (forall G A G' p v,
     expandp A p v
     -> wpof p A G'
     -> wvof (G;; G') v A).
Proof.
cut (forall N,
       (forall G r A m,
          size A N
          -> wrof G r (susp A)
          -> expandm A r m
          -> wtof G m A)
       /\
       (forall G A G' p v,
          size A N
          -> expandp A p v
          -> wpof p A G'
          -> wvof (G;; G') v A)).
  intro Hind.
  split.
    intros G r A m.
    so (ntp_size A) as (N & Hsize).
    revert G r A m Hsize.
    exact (Hind N andel).

    intros G A G' p v.
    so (ptp_size A) as (N & Hsize).
    revert G A G' p v Hsize.
    exact (Hind N ander).
intro N.
wfinduct N using lt_wf.
intros N IH.
so (fun N' Hlt => IH N' Hlt andel) as IHn.
so (fun N' Hlt => IH N' Hlt ander) as IHp.
clear IH.
split.



intros G r A m Hsize Hcompose Hexpand.
revert G Hsize Hcompose.
cases Hexpand.

(* atom *)
intros c s r G Hofr.
apply wtof_etam; auto; done.

(* up *)
intros A r p v r' HexpandA Hcompose G Hsize Hofr.   
invertc Hsize; [].
intros N' HsizeA <-.
so (can_pof_expandp_similar _#3 HexpandA) as (A' & G' & Hsim & Hofp).
so (expandp_similar _#4 Hsim HexpandA) as HexpandA'.
so (IHp _ (lt_n_Sn _) G _#4 (similar_size _#4 Hsim HsizeA) HexpandA' (pof_weak _#3 Hofp)) as Hofv.
apply wtof_shiftp; [].
eapply (wrof_similar _ _ (trp A')); eauto with similar; [].
#3 eapply rcompose_formation_weak.
  exact Hcompose.

  eapply (wrof_similar _ _ _ (susp (up A'))); eauto using @similar; done.
eauto using wsof, wrof, pof_weak; done.

(* pi *)
intros p' A B r p v r' m HexpandA Hcompose HexpandB G Hsize Hofr.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
so (can_pof_expandp_similar _#3 HexpandA) as (A' & G' & Hsim & Hofp).
so (expandp_similar _#4 Hsim HexpandA) as HexpandA'.
so (IHp _ (le_lt_n_Sm _ _ (le_plus_l _ _)) G _#4 (similar_size _#4 Hsim HsizeA) HexpandA' (pof_weak _#3 Hofp)) as Hofv.
#2 exploit (IHn _ (le_lt_n_Sm _ _ (le_plus_r _ _)) (G;; G') r' B m) as Hofm; auto.
  refine (rcompose_formation_weak _ (pi p A' B) _#4 Hcompose _ _); eauto using wsof; [].
  apply (wrof_similar _ _ (susp (pi p' A B))); eauto with similar; [].
  apply (weaken_wrof G G' nil); eauto using pof_bound_length; done.
apply (wtof_similar _ _ (pi p A' B)); eauto with similar; [].
eapply wtof_lam; eauto using pof_weak; done.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB G Hsize Hofr.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
#2 exploit (IHn _ (le_lt_n_Sm _ _ (le_plus_l _ _)) G r1 A m1) as Hofm1; auto.
  eapply rcompose_formation_weak; eauto using wsof; done.
#2 exploit (IHn _ (le_lt_n_Sm _ _ (le_plus_r _ _)) G r2 B m2) as Hofm2; auto.
  eapply rcompose_formation_weak; eauto using wsof; done.
eauto using wtof; done.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB G Hsize Hofr.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
sor (refine (wrof_similar _#3 (susp (wth A B)) Hofr _)) as Hofr'; eauto with similar; [].
#2 exploit (IHn _ (le_lt_n_Sm _ _ (le_plus_l _ _)) G r1 A m1) as Hofm1; auto.
  eapply rcompose_formation_weak; eauto using wsof; done.
#2 exploit (IHn _ (le_lt_n_Sm _ _ (le_plus_r _ _)) G r2 B m2) as Hofm2; auto.
  eapply rcompose_formation_weak; eauto using wsof; done.
eauto using wtof; done.




intros G A G' p v Hsize Hexpand Hofp.
revert G G' Hsize Hofp.
cases Hexpand.

(* down *)
intros A m HexpandA G G' Hsize Hofp.
invertc Hsize; [].
intros N' HsizeA <-.
invertc Hofp; [].
intros <-.
simpl.
#2 exploit (IHn _ (lt_n_Sn _) (G; A) (use (var 0) nl) A m); auto.
  apply (@wrof_use _#3 (sh 0 1 A) A); eauto using lookup_var, index_0, wsof; [].
  apply similar_symm; [].
  apply sh_similarm; auto; done.
apply wvof_thunk; auto; done.  

(* one *)
intros G G' Hsize Hofp.
apply wvof_unitp; done.

(* tens *)
intros A B p q v w HexpandA HexpandB G G' Hsize Hofpat.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invertc Hofpat; [].
intros G1 G2 Hofp Hofq <-.
autorewrite with canonlist.
exploit (IHp _ (le_lt_n_Sm _ _ (le_plus_l _ _)) G A G1 p v) as Hofv; eauto; [].
exploit (IHp _ (le_lt_n_Sm _ _ (le_plus_r _ _)) G B G2 q w) as Hofw; eauto; [].
apply (wvof_similarctx (G;; G1;; G2)); eauto using shctx_similarctx with similar; [].
#2 apply wvof_pairp.
  apply (weaken_wvof (G;; G1) G2 nil); eauto; [].
  eapply wpof_bound_length; eauto; done.

  rewrite <- (expandp_closed 0 (length G1) _#3 HexpandB); [].
  repl (bound q + 0) with (bound q) by omega.
  apply weaken_wvof; eauto using wpof_bound_length; done.

(* ex *)
intros p' A B p q v w HexpandA HexpandB G G' Hsize Hofpat.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invertc Hofpat; [].
intros G1 G2 Hofp Hofq <-.
autorewrite with canonlist.
exploit (IHp _ (le_lt_n_Sm _ _ (le_plus_l _ _)) G A G1 p v) as Hofv; eauto; [].
exploit (IHp _ (le_lt_n_Sm _ _ (le_plus_r _ _)) (G;; G1) B G2 q w) as Hofw; eauto; [].
#2 apply wvof_pairpx.
  apply (weaken_wvof (G;; G1) G2 nil); eauto using wpof_bound_length; done.

  exact Hofw.
Qed.


Lemma expandm_formation_weak :
  forall G r A m,
    wrof G r (susp A)
    -> expandm A r m
    -> wtof G m A.
Proof.
exact (identity_expansion_weak andel).
Qed.


Lemma expandp_formation_weak :
  forall G A G' p v,
    expandp A p v
    -> wpof p A G'
    -> wvof (G;; G') v A.
Proof.
exact (identity_expansion_weak ander).
Qed.
