
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Similar.
Require Import StaticWeak.
Require Import Substitution.
Require Import Weakening.
Require Import IdentityWeak.
Require Import CutEliminationWeak.


Lemma wpof_fun :
  forall A p p' G G',
    wpof p A G
    -> wpof p' A G'
    -> p = p' /\ G = G'.
Proof.
intros A p p' G G' H Hofp'.
revert p' G' Hofp'.
induct H.

(* down *)
intros A p' G' Hofp'.
invert Hofp'; [].
intros; subst; auto; done.

(* one *)
intros p' G' Hofp'.
invert Hofp'; [].
intros; subst; auto; done.

(* tens *)
intros p1 p2 A B G1 G2 _ IH1 _ IH2 pp GG Hofpp.
invert Hofpp; clear Hofpp; [].
intros p1' p2' G1' G2' Hofp1' Hofp2' <- <-.
so (IH1 _ _ Hofp1') as (<- & <-).
so (IH2 _ _ Hofp2') as (<- & <-).
auto; done.

(* ex *)
intros p1 p2 p1'' A B G1 G2 Hofp1 IH1 Hofp2 IH2 pp GG Hofpp.
invertc Hofpp; [].
intros p1' p2' G1' G2' Hofp1' Hofp2' <- <-.
so (IH1 _ _ Hofp1') as (<- & <-).
so (IH2 _ _ Hofp2') as (<- & <-).
auto.
Qed.


Lemma lookup_reg_weak :
  forall G h A,
    wwf nil G
    -> lookup G h A
    -> wntof G A.
Proof.
intros G h A Hwf Hlookup.
revert Hwf.
cases Hlookup.

(* const *)
intros G c A Hofc Hwf.
repl G with (nil;; G;; nil) at 1 by (autorewrite with canonlist; reflexivity).
apply weaken_wntof; auto; [].
simpl.
eapply ntof_weak; eauto using constof_ntof; done.

(* var *)
intros G i A Hindex Hwf.
revert Hwf.
#2 induct Hindex.
  (* 0 *)
  intros A G Hwf.
  invert Hwf; [].
  intros _ HofA.
  autorewrite with canonlist in HofA.
  apply weaken1_wntof; auto; done.

  (* S *)
  intros i B G A _ IH Hwf.
  invert Hwf; [].
  intros Hwf' _.
  so (IH Hwf') as HofA.
  so (weaken1_wntof _ B _ HofA) as HofA'.
  rewrite -> sh_sum in HofA'.
  exact HofA'.
Qed.


Lemma self_reg_weak :
  forall G m A B,
    self m A B
    -> wtof G m A
    -> wntof G A
    -> wntof G B.
Proof.
intros G m A B Hself Hofm Hof.
revert G Hofm Hof.
#3 induct Hself; auto.

(* pi *)
intros p m A B C _ IH G Hoflam Hof.
invertc Hof; [].
intros G' HofA Hofp HofB.
invertc Hoflam; [].
intros G'' Hofp' Hofm.
so (wpof_fun _#5 Hofp Hofp') as (_ & <-).
eauto using wntof; done.

(* wth *)
intros m n A B C D _ IH1 _ IH2 G Hofpair Hof.
invertc Hof; [].
intros HofA HofB.
invertc Hofpair; [].
intros Hofm Hofn.
eauto using wntof; done.

(* sigma *)
intros m n A B C B' D _ IH1 HsubB ? IH2 G Hofpair Hof.
invertc Hof; [].
intros HofA HofB.
invertc Hofpair; [].
intros Hofm Hofn.
apply wntof_wth; eauto; [].
#2 apply IH2.
  eapply wtof_similar; eauto using rsub_similarm; done.

  repl (G; A) with (G; A;; nil) in HofB by reflexivity.
  so (can_rsub_nt_weak _#6 (eq_refl _) Hofm HofB) as (B'' & HsubB' & HofB').
  so (rsub_fun _#5 HsubB HsubB'); subst B''.
  simpl in HofB'.
  exact HofB'.
Qed.


Lemma hof_reg_weak :
  forall G h A,
    wwf nil G
    -> hof G h A
    -> wntof G A.
Proof.
intros G h A Hwf Hofh.
revert Hwf.
cases Hofh.

(* id *)
intros G h A H Hwf.
eapply lookup_reg_weak; eauto; done.

(* self *)
intros G h m A B Hlookup Hexpand Hself Hwf.
eapply self_reg_weak; eauto using lookup_reg_weak; [].
eapply expandm_formation_weak; eauto; [].
eapply wrof_use; eauto using wsof with similar; done.
Qed.


Lemma sof_reg_weak :
  forall G s A B,
    wntof G A
    -> sof G s A (susp B)
    -> wntof G B.
Proof.
intros G s A B HofA Hofs.
remember (susp B) as U eqn:HeqU.
revert B HofA HeqU.
induct Hofs.

(* nl *)
intros G A B HofA HeqU.
injection HeqU; intros; subst; eauto; done.

(* pm *)
intros G p r A B G' _ _ C _ Heq.
discriminate Heq.

(* app *)
intros G v s p A B U B' Hofv HsubB Hofs IH C Hofty ->.
invertc Hofty; [].
intros G' HofA Hofp HofB.
apply IH; eauto; [].
repl (G;; G') with (G;; G';; nil) in HofB by reflexivity.
so (can_cutp_nt_weak _#8 (eq_refl _) Hofp (vof_weak _#3 Hofv) HofB) as (B'' & HsubB' & HofB').
so (cutp_fun _#6 HsubB HsubB'); subst B''.
simpl in HofB'.
exact HofB'.

(* pi1 *)
intros G s A B U _ IH C Hofty ->.
invertc Hofty; [].
intros HofA HofB.
eapply IH; eauto; done.

(* pi2 *)
intros G s A B U _ IH C Hofty ->.
invertc Hofty; [].
intros HofA HofB.
eapply IH; eauto; done.
Qed.


Lemma rof_reg_weak :
  forall G r A,
    wwf nil G
    -> rof G r (susp A)
    -> wntof G A.
Proof.
intros G r A Hwf Hofr.
invert Hofr; [].
intros h s B Hofh Hofs <-.
eapply sof_reg_weak; eauto; [].
eapply hof_reg_weak; eauto; done.
Qed.


Lemma wwf_append :
  forall G G1 G2,
    wwf G G1
    -> wwf (G;; G1) G2
    -> wwf G (G1;; G2).
Proof.
intros G G1 G2 Hwf1 Hwf2.
revert G G1 Hwf1 Hwf2.
induct G2.

(* nil *)
intros G G1 Hwf1 Hwf2.
simpl.
exact Hwf1.

(* cons *)
intros A G2 H G G1 Hwf1 Hwf.
invertc Hwf; [].
intros Hwf2 HofA.
simpl.
apply wwf_cons; auto; [].
rewrite <- app_assoc; [].
exact HofA.
Qed.


Lemma wwf_append_closed :
  forall G1 G2,
    wwf nil G1
    -> wwf G1 G2
    -> wwf nil (G1;; G2).
Proof.
intros G1 G2 Hwf1 Hwf2.
repl G1 with (nil;; G1) in Hwf2 by (autorewrite with canonlist; reflexivity).
apply wwf_append; auto.
Qed.


Lemma pof_reg_weak :
  forall G p A G',
    wptof G A
    -> pof p A G'
    -> wwf G G'.
Proof.
intros G p A G' Hofty Hofp.
revert G Hofty.
induct Hofp.

(* down *)
intros A G Hofty.
invertc Hofty; [].
eauto using wwf; done.

(* one *)
intros G Hofty.
exact wwf_nil.

(* tens *)
intros p q A B G2 G3 Hofp IH1 Hofq IH2 G1 Hofty.
invertc Hofty; [].
intros HofA HofB.
apply wwf_append; eauto; [].
repl (G1;; G2) with (G1;; G2;; nil) by reflexivity.
eapply weaken_wwf; eauto; done.

(* ex *)
intros p q A B G2 G3 Hofp IH1 Hofq IH2 G1 Hofty.
invertc Hofty; [].
intros G2' HofA Hofp' HofB.
so (wpof_fun _#5 (pof_weak _#3 Hofp) Hofp') as (_ & <-).
apply wwf_append; eauto; done.
Qed.
