
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.


Inductive compose : forall {z} (* sp/re *), exp z -> spine -> exp z -> Prop :=
| compose_nl {s}
    : compose nl s s

| compose_pm {p r1 s2 r}
    : compose r1 (sh 0 (bound p) s2) r
      -> compose (pm p r1) s2 (pm p r)

| compose_app {v s1 s2 s}
    : compose s1 s2 s
      -> compose (app v s1) s2 (app v s)

| compose_pi1 {s1 s2 s}
    : compose s1 s2 s
      -> compose (pi1 s1) s2 (pi1 s)

| compose_pi2 {s1 s2 s}
    : compose s1 s2 s
      -> compose (pi2 s1) s2 (pi2 s)

| compose_use {h s1 s2 s}
    : compose s1 s2 s
      -> compose (use h s1) s2 (use h s)
.


Inductive expandm : ntp -> res -> term -> Prop :=
| expandm_atom {c s r}
    : expandm (tuse c s) r (etam r)

| expandm_up {A r p v r'}
    : expandp A p v
      -> compose r (pm p (ret v)) r'
      -> expandm (up A) r (shiftp r')

| expandm_pi {p' A B r p v r' m}
    : expandp A p v
      -> compose (sh 0 (bound p) r) (app v nl) r'
      -> expandm B r' m
      -> expandm (pi p' A B) r (lam p m)

| expandm_wth {A B r r1 r2 m1 m2}
    : compose r (pi1 nl) r1
      -> compose r (pi2 nl) r2
      -> expandm A r1 m1
      -> expandm B r2 m2
      -> expandm (wth A B) r (pairm m1 m2)

| expandm_sigma {A B r r1 r2 m1 m2}
    : compose r (pi1 nl) r1
      -> compose r (pi2 nl) r2
      -> expandm A r1 m1
      -> expandm B r2 m2
      -> expandm (sigma A B) r (pairm m1 m2)


with expandp : ptp -> pattern -> val -> Prop :=
| expandp_down {A m}
    : expandm A (use (var 0) nl) m
      -> expandp (down A) shiftm (thunk m)

| expandp_one
    : expandp one unitpat unitp

| expandp_tens {A B p q v w}
    : expandp A p v
      -> expandp B q w
      -> expandp (tens A B) (pairpat p q) (pairp (sh 0 (bound q) v) w)

| expandp_ex {p' A B p q v w}
    : expandp A p v
      -> expandp B q w
      -> expandp (ex p' A B) (pairpat p q) (pairp (sh 0 (bound q) v) w)
.


Scheme expandm_mut_ind := Minimality for expandm Sort Prop
with   expandp_mut_ind := Minimality for expandp Sort Prop.
Combined Scheme expand_mut_ind from expandm_mut_ind, expandp_mut_ind.


Inductive self : term -> ntp -> ntp -> Prop :=
| self_atom {m c s}
    : self m (tuse c s) (tuse c s)

| self_up {m A}
    : self m (up A) (up A)

| self_pi {p m A B B'}
    : self m B B'
      -> self (lam p m) (pi p A B) (pi p A B')

| self_wth {m n A B A' B'}
    : self m A A'
      -> self n B B'
      -> self (pairm m n) (wth A B) (wth A' B')

| self_sigma {m n A B A' B' B''}
    : self m A A'
      -> rsub m 0 B B'
      -> self n B' B''
      -> self (pairm m n) (sigma A B) (wth A' B'')
.
