
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Self.
Require Import Cut.
Require Import Static.
Require Import Similar.


Inductive wpof : pattern -> ptp -> context -> Prop :=
| wpof_shiftm {A}
    : wpof shiftm (down A) (nil ; A)

| wpof_unitpat
    : wpof unitpat one nil

| wpof_pairpat {p p' A A' G G'}
    : wpof p A G
      -> wpof p' A' G'
      -> wpof (pairpat p p') (tens A A') (G ;; shctx 0 (length G) G')

| wpof_pairpatx {p p' p'' A B G G'}
    : wpof p A G
      -> wpof p' B G'
      -> wpof (pairpat p p') (ex p'' A B) (G ;; G')
.


Inductive wrof : context -> res -> con -> Prop :=
| wrof_ret {G v A}
    : wvof G v A
      -> wrof G (ret v) (trp A)

| wrof_use {G h s A A' U}
    : lookup G h A
      -> similar A A'
      -> wsof G s A' U
      -> wrof G (use h s) U

with wvof : context -> val -> ptp -> Prop :=
| wvof_thunk {G m A}
    : wtof G m A
      -> wvof G (thunk m) (down A)

| wvof_unitp {G}
    : wvof G unitp one

| wvof_pairp {G v w A B}
    : wvof G v A
      -> wvof G w B
      -> wvof G (pairp v w) (tens A B)

| wvof_pairpx {G v w p A B}
    : wvof G v A
      -> wvof G w B
      -> wvof G (pairp v w) (ex p A B)

with wsof : context -> spine -> ntp -> con -> Prop :=
| wsof_nl {G A}
    : wsof G nl A (susp A)

| wsof_pm {G p r A B G'}
    : wpof p A G'
      -> wrof (G ;; G') r (trp B)
      -> wsof G (pm p r) (up A) (trp B)

| wsof_app {G v s p A B U}
    : wvof G v A
      -> wsof G s B U
      -> wsof G (app v s) (pi p A B) U

| wsof_pi1 {G s A B U}
    : wsof G s A U
      -> wsof G (pi1 s) (wth A B) U

| wsof_pi2 {G s A B U}
    : wsof G s B U
      -> wsof G (pi2 s) (wth A B) U


with wtof : context -> term -> ntp -> Prop :=
| wtof_etam {G r h s}
      : wrof G r (susp (tuse h s))
      -> wtof G (etam r) (tuse h s)

| wtof_shiftp {G r A}
    : wrof G r (trp A)
      -> wtof G (shiftp r) (up A)

| wtof_lam {G p m p' A B G'}
    : wpof p A G'
      -> wtof (G ;; G') m B
      -> wtof G (lam p m) (pi p' A B)

| wtof_pairm {G m n A B}
    : wtof G m A
      -> wtof G n B
      -> wtof G (pairm m n) (wth A B)

| wtof_pairms {G m n A B}
    : wtof G m A
      -> wtof G n B
      -> wtof G (pairm m n) (sigma A B)
.


Inductive wtsof : context -> tspine -> kind -> kind -> Prop :=
| wtsof_nl {G K}
    : wtsof G tnl K K

| wtsof_app {G v s p A K U}
    : wvof G v A
      -> wtsof G s K U
      -> wtsof G (tapp v s) (tpi p A K) U
.


Inductive wntof : context -> ntp -> Prop :=
| wntof_tuse {G c s K K'}
    : tconstof c K
      -> similar K K'
      -> wtsof G s K' ty
      -> wntof G (tuse c s)

| wntof_up {G A}
    : wptof G A
      -> wntof G (up A)

| wntof_pi {G p A B G'}
    : wptof G A
      -> wpof p A G'
      -> wntof (G;; G') B
      -> wntof G (pi p A B)

| wntof_wth {G A B}
    : wntof G A
      -> wntof G B
      -> wntof G (wth A B)

| wntof_sigma {G A B}
    : wntof G A
      -> wntof (G; A) B
      -> wntof G (sigma A B)

with wptof : context -> ptp -> Prop :=
| wptof_down {G A}
    : wntof G A
      -> wptof G (down A)

| wptof_one {G}
    : wptof G one

| wptof_tens {G A B}
    : wptof G A
      -> wptof G B
      -> wptof G (tens A B)

| wptof_ex {G p A B G'}
    : wptof G A
      -> wpof p A G'
      -> wptof (G;; G') B
      -> wptof G (ex p A B)
.


Inductive wkof : context -> kind -> Prop :=
| wkof_ty {G}
    : wkof G ty

| wkof_tpi {G p A K G'}
    : wptof G A
      -> wpof p A G'
      -> wkof (G;; G') K
      -> wkof G (tpi p A K)
.


Scheme wrof_mut_ind := Minimality for wrof Sort Prop
with   wvof_mut_ind := Minimality for wvof Sort Prop
with   wsof_mut_ind := Minimality for wsof Sort Prop
with   wtof_mut_ind := Minimality for wtof Sort Prop.
Combined Scheme weak_static_term_ind from wrof_mut_ind, wvof_mut_ind, wsof_mut_ind, wtof_mut_ind.

Scheme wntof_mut_ind := Minimality for wntof Sort Prop
with   wptof_mut_ind := Minimality for wptof Sort Prop.
Combined Scheme weak_static_tp_ind from wntof_mut_ind, wptof_mut_ind.



Inductive wwf : context -> context -> Prop :=
| wwf_nil {G}
    : wwf G nil

| wwf_cons {G G' A}
    : wwf G G'
      -> wntof (G;; G') A
      -> wwf G (G'; A)
.


Inductive wgof : forall {z}, context -> exp z -> judgement -> Prop :=
| wgof_r {G r U}
    : wrof G r U
      -> suspnormal U
      -> wgof G r (judr U)

| wgof_v {G v A}
    : wvof G v A
      -> wgof G v (judv A)

| wgof_s {G s A U}
    : wsof G s A U
      -> suspnormal U
      -> wgof G s (juds A U)

| wgof_t {G m A}
    : wtof G m A
      -> wgof G m (judt A)

| wgof_ts {G s K K'}
    : wtsof G s K K'
      -> wgof G s (judts K K')

| wgof_nt {G A}
    : wntof G A
      -> wgof G A judnt

| wgof_pt {G A}
    : wptof G A
      -> wgof G A judpt

| wgof_k {G A}
    : wkof G A
      -> wgof G A judk
.


Lemma similarctx_index :
  forall j G G' A,
    similarctx G G'
    -> index j G A
    -> exists A',
         similar A A'
         /\ index j G' A'.
Proof.
intros j G G' A Hsim Hindex.
revert G' Hsim.
induct Hindex.

(* 0 *)
intros A G GG Hsim.
invert Hsim; [].
intros A' G' HsimA HsimG <-.
exists A'.
split; eauto using index_0; done.

(* S *)
intros j B G A _ IH GG Hsim.
invert Hsim; [].
intros B' G' _ HsimG <-.
so (IH _ HsimG) as (A' & HsimA & Hindex).
exists A'.
split; eauto using index_S; done.
Qed.


Lemma lookup_similarctx :
  forall G G' h A,
    lookup G h A
    -> similarctx G G'
    -> exists A',
         similar A A'
         /\ lookup G' h A'.
Proof.
intros G G' h A Hlookup HsimG.
#2 invertc Hlookup.
  intros c B Hofc <- <-.
  exists (sh 0 (length G) B).
  rewrite -> (similarctx_length _ _ HsimG); [].
  split; auto using lookup_const with similar; done.

  intros i B Hindex <- <-.
  so (similarctx_index _#4 HsimG Hindex) as (B' & HsimB & Hindex').
  exists (sh 0 (S i) B').
  split; eauto using lookup_var; [].
  #2 eapply similar_trans.
    apply similar_symm; apply sh_similarm; eauto; done.

    eapply similar_trans; eauto using sh_similarm; done.
Qed.


Lemma wpof_similar :
  forall p A A' G,
    wpof p A G
    -> similar A A'
    -> exists G',
         similarctx G G'
         /\ wpof p A' G'.
Proof.
intros p A A' G Hofp Hsim.
revert A' Hsim.
induct Hofp.

(* down *)
intros A AA Hsim.
invertc Hsim; [].
intros A' HsimA <-.
exists (nil; A').
eauto using wpof with similar; done.

(* one *)
intros AA Hsim.
invertc Hsim; [].
intros <-.
exists nil; eauto using wpof with similar; done.

(* tens *)
intros p q A B G1 G2 _ IH1 _ IH2 AA Hsim.
#2 invertc Hsim.
  intros A' B' HsimA HsimB <-.
  so (IH1 _ HsimA) as (G1' & HsimG1 & Hofp).
  so (IH2 _ HsimB) as (G2' & HsimG2 & Hofq).
  exists (G1';; shctx 0 (length G1') G2'); split; eauto using wpof; [].
  apply similarctx_append; auto; [].
  rewrite <- (similarctx_length _ _ HsimG1); [].
  #2 eapply similarctx_trans.
    apply similarctx_symm; apply shctx_similarctx; done.

    eapply similarctx_trans; eauto using shctx_similarctx; done.

  intros p' A' B' HsimA HsimB <-.
  so (IH1 _ HsimA) as (G1' & HsimG1 & Hofp).
  so (IH2 _ HsimB) as (G2' & HsimG2 & Hofq).
  exists (G1';; G2'); split; eauto using wpof; [].
  apply similarctx_append; auto; [].
  eapply similarctx_trans; eauto; [].
  apply similarctx_symm; apply shctx_similarctx; done.

(* ex *)
intros p q p' A B G1 G2 _ IH1 _ IH2 AA Hsim.
#2 invertc Hsim.
  intros p'' A' B' HsimA HsimB <-.
  so (IH1 _ HsimA) as (G1' & HsimG1 & Hofp).
  so (IH2 _ HsimB) as (G2' & HsimG2 & Hofq).
  exists (G1';; G2'); split; eauto using wpof; [].
  apply similarctx_append; auto; done.

  intros A' B' HsimA HsimB <-.
  so (IH1 _ HsimA) as (G1' & HsimG1 & Hofp).
  so (IH2 _ HsimB) as (G2' & HsimG2 & Hofq).
  exists (G1';; shctx 0 (length G1') G2'); split; eauto using wpof; [].
  apply similarctx_append; auto; [].
  eapply similarctx_trans; eauto; [].
  apply shctx_similarctx; done.
Qed.


Lemma similarity :
  (forall G G' r U U',
     similarctx G G'
     -> similar U U'
     -> wrof G r U
     -> wrof G' r U')
  /\
  (forall G G' v A A',
     similarctx G G'
     -> similar A A'
     -> wvof G v A
     -> wvof G' v A')
  /\
  (forall G G' s A U U',
     similarctx G G'
     -> similar U U'
     -> wsof G s A U
     -> exists A',
          similar A A'
          /\ wsof G' s A' U')
  /\
  (forall G G' m A A',
     similarctx G G'
     -> similar A A'
     -> wtof G m A
     -> wtof G' m A').
Proof.
#7 exploit
  (weak_static_term_ind
     (fun G r U =>
        forall G' U',
          similarctx G G'
          -> similar U U'
          -> wrof G' r U')
     (fun G v A =>
        forall G' A',
          similarctx G G'
          -> similar A A'
          -> wvof G' v A')
     (fun G s A U =>
        forall G' U',
          similarctx G G'
          -> similar U U'
          -> exists A',
               similar A A'
               /\ wsof G' s A' U')
     (fun G m A =>
        forall G' A',
          similarctx G G'
          -> similar A A'
          -> wtof G' m A'))
  as Hind;
try (intros;
     try (match goal with
          | H : similar _ _ |- _ => invertc H
          end);
     intros; subst;
     eauto using wrof, wvof, wsof, wtof with similar;
     done).

(* use *)
intros G h s A A1 U Hlookup HsimA1 _ IH G' U' HsimG HsimU.
so (IH _ _ HsimG HsimU) as (A2 & HsimA2 & Hsof).
so (lookup_similarctx _#4 Hlookup HsimG) as (A3 & HsimA3 & Hlookup').
eapply wrof_use; eauto; [].
apply (similar_trans _ A1); eauto with similar; done.

(* pm *)
intros G1 p r A B G2 Hofp Hofr IH G1' U' HsimG1 HsimU.
invertc HsimU; [].
intros B' HsimB <-.
exists (up A).
split; eauto using wsof with similar; done.

(* app *)
intros G v s p A B U _ IH1 _ IH2 G' U' HsimG HsimU.
so (IH1 _ _ HsimG (similarp_refl A)) as Hofv.
so (IH2 _ _ HsimG HsimU) as (B' & HsimB & Hofs).
exists (pi p A B').
split; eauto using wsof with similar; done.

(* pi1 *)
intros G s A B U _ IH G' U' HsimG HsimU.
so (IH _ _ HsimG HsimU) as (A' & HsimA & Hofs).
exists (wth A' B).
split; eauto using wsof with similar; done.

(* pi2 *)
intros G s A B U _ IH G' U' HsimG HsimU.
so (IH _ _ HsimG HsimU) as (B' & HsimB & Hofs).
exists (wth A B').
split; eauto using wsof with similar; done.

(* lam *)
intros G1 p m p' A B G2 Hofp _ IH G1' AA HsimG1 Hsim.
invertc Hsim; [].
intros p'' A' B' HsimA HsimB <-.
so (wpof_similar _#4 Hofp HsimA) as (G2' & HsimG2 & Hofp').
exploit (IH (G1';; G2') B') as Hofm; eauto with similar; [].
eapply wtof_lam; eauto; done.

(* wrapup *)
destruct_all Hind.
do2 3 split; intros; eauto.
Qed.


Lemma wrof_similar :
  forall G r U U',
    wrof G r U
    -> similar U U'
    -> wrof G r U'.
Proof.
intros.
eapply (similarity andel); eauto with similar.
Qed.


Lemma wvof_similar :
  forall G v A A',
    wvof G v A
    -> similar A A'
    -> wvof G v A'.
Proof.
intros.
eapply (similarity anderl); eauto with similar.
Qed.


Lemma wtof_similar :
  forall G m A A',
    wtof G m A
    -> similar A A'
    -> wtof G m A'.
Proof.
intros.
eapply (similarity anderrr); eauto with similar.
Qed.


Lemma wrof_similarctx :
  forall G G' r U,
    wrof G r U
    -> similarctx G G'
    -> wrof G' r U.
Proof.
intros.
eapply (similarity andel); eauto with similar.
Qed.


Lemma wvof_similarctx :
  forall G G' v A,
    wvof G v A
    -> similarctx G G'
    -> wvof G' v A.
Proof.
intros.
eapply (similarity anderl); eauto with similar.
Qed.


Lemma wsof_similarctx :
  forall G G' s A U,
    wsof G s A U
    -> similarctx G G'
    -> wsof G' s A U.
Proof.
intros G G' s A U Hofs Hsim.
revert G' Hsim.
induct Hofs; intros; eauto using wrof_similarctx, wvof_similarctx, wsof with similar.
Qed.


Lemma wtof_similarctx :
  forall G G' m A,
    wtof G m A
    -> similarctx G G'
    -> wtof G' m A.
Proof.
intros.
eapply (similarity anderrr); eauto with similar.
Qed.


Lemma wtsof_similarctx :
  forall G G' s K L,
    wtsof G s K L
    -> similarctx G G'
    -> wtsof G' s K L.
Proof.
intros G G' s K L Hofs Hsim.
revert G' Hsim.
induct Hofs.

(* nl *)
eauto using wtsof; done.

(* app *)
intros G v s p A K U Hofv _ IH G' Hsim.
so (wvof_similarctx _#4 Hofv Hsim) as Hofv'.
so (IH _ Hsim) as Hofs'.
eauto using wtsof.
Qed.


Lemma wtpof_similarctx :
  (forall G G' A,
     wntof G A
     -> similarctx G G'
     -> wntof G' A)
  /\
  (forall G G' A,
     wptof G A
     -> similarctx G G'
     -> wptof G' A).
Proof.
#4 exploit
  (weak_static_tp_ind
     (fun G A => forall G', similarctx G G' -> wntof G' A)
     (fun G A => forall G', similarctx G G' -> wptof G' A))
  as Hind;
  try (intros; eauto using wntof, wptof, wtsof_similarctx; done).

(* pi *)
intros G1 p A B G2 _ IH1 Hofp _ IH2 G1' Hsim.
so (IH1 _ Hsim) as HofA.
exploit (IH2 (G1';; G2)) as HofB; eauto with similar; [].
eauto using wntof; done.

(* sigma *)
intros G A B _ IH1 _ IH2 G' Hsim.
so (IH1 _ Hsim) as HofA.
exploit (IH2 (G'; A)) as HofB'; eauto with similar; [].
eauto using wntof; done.

(* ex *)
intros G1 p A B G2 _ IH1 Hofp _ IH2 G1' Hsim.
so (IH1 _ Hsim) as HofA.
exploit (IH2 (G1';; G2)) as HofB; eauto with similar; [].
eauto using wptof; done.

(* wrapup *)
destruct Hind; split; intros; eauto.
Qed.


Lemma wntof_similarctx :
  forall G G' A,
    wntof G A
    -> similarctx G G'
    -> wntof G' A.
Proof.
exact (wtpof_similarctx andel).
Qed.


Lemma wptof_similarctx :
  forall G G' A,
    wptof G A
    -> similarctx G G'
    -> wptof G' A.
Proof.
exact (wtpof_similarctx ander).
Qed.


Lemma wkof_similarctx :
  forall G G' K,
    wkof G K
    -> similarctx G G'
    -> wkof G' K.
Proof.
intros G G' K Hof Hsim.
revert G' Hsim.
induct Hof.

(* ty *)
eauto using wkof; done.

(* tpi *)
intros G1 p A K G2 HofA Hofp _ IH G1' Hsim.
so (wptof_similarctx _#3 HofA Hsim) as HofA'.
exploit (IH (G1';; G2)) as HofK; eauto with similar; [].
eauto using wkof; done.
Qed.


Lemma wgof_similarctx :
  forall {z} G G' (e : exp z) J,
    wgof G e J 
    -> similarctx G G'
    -> wgof G' e J.
Proof.
intros z G G' e J Hgof Hsim.
revert G' Hsim.
cases Hgof; intros; eauto using @wgof, wrof_similarctx, wvof_similarctx, wsof_similarctx, wtof_similarctx, wtsof_similarctx, wntof_similarctx, wptof_similarctx, wkof_similarctx.
Qed.


Lemma pof_weak :
  forall p A G,
    pof p A G
    -> wpof p A G.
Proof.
intros p A G H.
induct H; eauto using wpof.
Qed.


Lemma stat_weak :
  (forall G r U,
     rof G r U
     -> wrof G r U)
  /\
  (forall G v A,
     vof G v A
     -> wvof G v A)
  /\
  (forall G s A U,
     sof G s A U
     -> exists A',
          similar A A'
          /\ wsof G s A' U)
  /\
  (forall G m A,
     tof G m A
     -> wtof G m A).
Proof.
#8 exploit
  (static_term_ind
     (fun G r U => wrof G r U)
     (fun G v A => wvof G v A)
     (fun G s A U => exists A', similar A A' /\ wsof G s A' U)
     (fun G m A => wtof G m A))
  as Hind;
  eauto using wrof, wvof, wsof, wtof, pof_weak.

(* use *)
intros G h s U A Hofh _ IH.
destruct IH as (A' & Hsim & Hofs).
#2 invert Hofh.
  intros Hlookup.
  eauto using wrof with similar; done.

  intros m B Hlookup Hexpand Hself.
  eapply wrof_use; eauto; [].
  eapply similar_trans; eauto using self_similar; done.

(* pairp *)
intros G v w p A B B' _ Hofv HsubB _ Hofw.
apply wvof_pairpx; eauto; [].
eapply wvof_similar; eauto using cutp_similarp with similar; done.

(* nl *)
intros G A.
exists A; eauto using wsof with similar; done.

(* pm *)
intros G p r A B G' Hofp _ Hofr.
exists (up A); split; eauto with similar; [].
eapply wsof_pm; eauto using pof_weak; [].
eapply wrof_similar; eauto using sh_similarp with similar; done.

(* app *)
intros G v s p A B U B' _ Hofv HsubB _ IH.
destruct IH as (B'' & HsimB & Hofs).
exists (pi p A B'').
split; eauto using wsof; [].
apply similar_pi; eauto with similar; [].
apply (similar_trans _ B'); eauto using cutp_similarm with similar; done.

(* pi1 *)
intros G s A B U _ IH.
destruct IH as (A' & Hsim & Hofs).
exists (wth A' B); eauto using wsof with similar; done.

(* pi2 *)
intros G s A B U _ IH.
destruct IH as (B' & Hsim & Hofs).
exists (wth A B'); eauto using wsof with similar; done.

(* pairms *)
intros G m n A B B' _ Hofm HsubB _ Hofn.
apply wtof_pairms; auto; [].
eapply wtof_similar; eauto using rsub_similarm with similar; done.
Qed.


Lemma rof_weak :
  forall G r U,
    rof G r U
    -> wrof G r U.
Proof.
exact (stat_weak andel).
Qed.


Lemma vof_weak :
  forall G v A,
    vof G v A
    -> wvof G v A.
Proof.
exact (stat_weak anderl).
Qed.


Lemma tof_weak :
  forall G m A,
    tof G m A
    -> wtof G m A.
Proof.
exact (stat_weak anderrr).
Qed.


Lemma tsof_weak :
  forall G s K L,
    tsof G s K L
    -> exists K',
         similar K K'
         /\ wtsof G s K' L.
Proof.
intros G s K L H.
induct H.

(* nl *)
intros G K.
exists K; eauto using wtsof with similar; done.

(* app *)
intros G v s p A K U K' Hofv HsubK _ IH.
destruct IH as (K'' & HsimK & Hofs).
exists (tpi p A K'').
split; eauto using wtsof, vof_weak; [].
apply similar_tpi; eauto with similar; [].
apply (similar_trans _ K'); eauto using cutp_similarkd with similar; done.
Qed.


Lemma tpof_weak :
  (forall G A,
     ntof G A
     -> wntof G A)
  /\
  (forall G A,
     ptof G A
     -> wptof G A).
Proof.
exploit
  (static_tp_ind
     (fun G A => wntof G A)
     (fun G A => wptof G A))
  as Hind;
eauto using wntof, wptof, pof_weak; [].

(* use *)
intros G c s K Hofc Hofs.
so (tsof_weak _#4 Hofs) as (K' & Hsim & Hofs').
eapply wntof_tuse; eauto; [].
eapply similar_trans; eauto; [].
apply sh_similarkd; auto; done.
Qed.


Lemma ntof_weak :
  forall G A,
    ntof G A
    -> wntof G A.
Proof.
exact (tpof_weak andel).
Qed.
