
Require Import Arith.
Require Import Tactics.
Require Import Compare_dec.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Fuel.


(* Patmatch *)

Inductive patmatch : val -> pattern -> Prop :=
| patmatch_down {m}
    : patmatch (thunk m) shiftm

| patmatch_one
    : patmatch unitp unitpat

| patmatch_tens {v w p q}
    : patmatch v p
      -> patmatch w q
      -> patmatch (pairp v w) (pairpat p q)
.


Lemma sh_patmatch :
  forall i j v p,
    patmatch v p
    -> patmatch (sh i j v) p.
Proof.
intros i j v p H.
induct H; intros; simpl; eauto using patmatch.
Qed.


Lemma cutpf_patmatch :
   forall {z} a v p i (e e' : exp z),
     cutpf a v p i e e'
     -> patmatch v p.
Proof.
intros z a v p i e e' H.
induct H; intros; eauto using patmatch.
Qed.


Lemma cutp_patmatch :
   forall {z} v p i (e e' : exp z),
     cutp v p i e e'
     -> patmatch v p.
Proof.
intros z v p i e e' H.
induct H; intros; eauto using patmatch.
Qed.


Lemma patmatch_fun :
  forall v p p',
    patmatch v p
    -> patmatch v p'
    -> p = p'.
Proof.
intros v p p' H Hmatch.
revert p' Hmatch.
induct H; intros; invert Hmatch; intros; subst; f_equal; eauto.
Qed.


(* Mundane *)

Lemma shctx_length :
  forall i j G,
    length (shctx i j G) = length G.
Proof.
intros i j G.
revert i j.
induct G;
intros; simpl; eauto; f_equal; eauto.
Qed.


Lemma rsubctx_length :
  forall f i G G',
    rsubctx f i G G'
    -> length G = length G'.
Proof.
intros f i G G' H.
induct H; intros; simpl; eauto; f_equal; eauto.
Qed.


Lemma cutpctx_length :
  forall v p i G G',
    cutpctx v p i G G'
    -> length G = length G'.
Proof.
intros v p i G G' Hcut.
induct Hcut; intros; subst; eauto using rsubctx_length; etransitivity; eauto.
Qed.


Lemma shctx_index :
  forall i j k G A,
    index k G A
    -> index k (shctx i j G) (sh (i + length G - k - 1) j A).
Proof.
intros i j k G A H.
induct H.

(* 0 *)
intros A G.
simpl.
repl (i + S (length G) - 0 - 1) with (i + length G) by omega.
apply index_0; done.

(* S *)
intros k B G A Hindex IH.
simpl.
repl (i + S (length G) - S k - 1) with (i + length G - k - 1) by omega.
apply index_S; auto; done.
Qed.


Lemma rsubctx_index :
  forall f i j G G' A,
    rsubctx f i G G'
    -> index j G A
    -> exists A',
         rsub f (i + length G - j - 1) A A'
         /\ index j G' A'.
Proof.
intros f i j G G' A Hrsub Hindex.
revert f i G' Hrsub.
induct Hindex.

(* 0 *)
intros A G f i GG Hrsub.
invert Hrsub; [].
intros G' A' HrsubG HrsubA <-.
exists A'.
split.
  force_exact HrsubA.
  f_equal; [].
  simpl length.
  omega.

  apply index_0; done.

(* S *)
intros j B G A _ IH f i GG Hrsub.
invert Hrsub; [].
intros G' B' HrsubG _ <-.
so (IH _#3 HrsubG) as (A' & HrsubA & Hindex).
exists A'.
split.
  force_exact HrsubA.
  f_equal; [].
  simpl length.
  omega.

  apply index_S; auto; done.
Qed.


(* Trivial *)

Lemma sh_var_at_zero :
  forall i j,
    sh_var 0 i j = i + j.
Proof.
intros i j.
unfold sh_var.
forget (lt_dec j 0) as b; destruct b; omega.
Qed.


Lemma sh_head_zero :
  forall i h,
    sh_head i 0 h = h.
Proof.
intros i h.
destruct h as [c | j].
  reflexivity.

  simpl.
  f_equal; [].
  unfold sh_var.
  simpl.
  forget (lt_dec j i) as b.
  destruct b; auto; done.
Qed.


Lemma sh_zero :
  forall {z} i (e : exp z),
    sh i 0 e = e.
Proof.
intros z i e.
revert i.
induct e;
try (intros; simpl; f_equal; auto; done); [].
(* use *)
intros h e IH i.
simpl.
f_equal; auto; [].
destruct h as [c | j].
  reflexivity.

  simpl.
  f_equal; [].
  unfold sh_var.
  simpl.
  forget (lt_dec j i) as b.
  destruct b; auto; done.
Qed.


(* Functionality *)

Lemma rsub_head_fun :
  forall i h h' h'',
    rsub_head i h h'
    -> rsub_head i h h''
    -> h' = h''.
Proof.
intros i h h1 h2 Hsubh1 Hsubh2.
#3 invert Hsubh1; invert Hsubh2; try (intros; subst; eauto; discriminate).
  intros k ? <- <- k' ? Heq <-.
  injection Heq; intros; subst k'.
  exfalso; omega.

  intros k ? <- <- k' ? Heq <-.
  injection Heq; intros; subst k'.
  exfalso; omega.

  intros k ? <- <- k' ? Heq <-.
  injection Heq; intros; subst k'.
  reflexivity.
Qed.


Lemma rsub_head_contra :
  forall i h,
    rsub_head i (var i) h
    -> False.
Proof.
intros i h H.
invert H; intros; omega.
Qed.


Lemma sub_fun :
  (forall z v p i (e e1 e2 : exp z),
     cutp v p i e e1
     -> cutp v p i e e2
     -> e1 = e2)
  /\
  (forall m s r r',
     cutm m s r
     -> cutm m s r'
     -> r = r')
  /\
  (forall z f i (e e' e'' : exp z),
     rsub f i e e'
     -> rsub f i e e''
     -> e' = e'')
  /\
  (forall z (e : exp z) p r e1 e2,
     lsub e p r e1
     -> lsub e p r e2
     -> e1 = e2).
Proof.
#5 exploit
  (substitution_ind
     (fun z v p i (E E1 : exp z) => forall E2 (HH : cutp v p i E E2), E1 = E2)
     (fun m s r1 => forall r2 (HH : cutm m s r2), r1 = r2)
     (fun z f i (E E1 : exp z) => forall E2 (HH : rsub f i E E2), E1 = E2)
     (fun z (E : exp z) p r E1 => forall E2 (HH : lsub E p r E2), E1 = E2))
  as Hind;
try (intros; invert HH; intros; subst; eauto; f_equal; eauto; done).

(* tens *)
intros z va v' p p' i e1 e2 e3 _ IH1 _ IH2 e3' HH.
invert HH; clear HH; [].
intros e2' Hcut1 Hcut2.
so (IH1 _ Hcut1); subst e2'.
so (IH2 _ Hcut2); subst e3'.
eauto; done.

(* pi *)
intros p m v s r m' _ IH1 _ IH2 r2 HH.
invert HH; clear HH; [].
intros m2 Hcutp2 Hcutm2.
so (IH1 _ Hcutp2); subst m2.
so (IH2 _ Hcutm2); subst r2.
reflexivity.

(* inst *)
intros f i h s h1 s1 Hsubh1 _ IH E2 HH.
#2 invertc HH.
  intros; subst; f_equal; eauto using rsub_head_fun; done.

  intros; subst; exfalso; eapply rsub_head_contra; eauto; done.

(* reduce *)
intros f i s r1 s1 _ IH1 _ IH2 r2 HH.
#2 invertc HH.
  intros; subst; exfalso; eapply rsub_head_contra; eauto; done.

  intros s2 Hsubs2 Hcut2.
  so (IH1 _ Hsubs2); subst s2.
  so (IH2 _ Hcut2); subst r2.
  reflexivity.

(* wrapup *)
destruct_all Hind.
do2 3 split; eauto.
Qed.


Lemma rsub_fun :
  forall {z} f i (e e' e'' : exp z),
    rsub f i e e'
    -> rsub f i e e''
    -> e' = e''.
Proof.
exact (sub_fun anderrl).
Qed.


Lemma cutp_fun :
  forall {z} v p i (e e' e'' : exp z),
    cutp v p i e e'
    -> cutp v p i e e''
    -> e' = e''.
Proof.
exact (sub_fun andel).
Qed.


(* Substitution/shifting commutativity *)

Lemma sh_sum :
  forall {z} i j j' (e : exp z),
    sh i j' (sh i j e) = sh i (j'+j) e.
Proof.
intros z i j j' e.
revert i.
induct e; try (intros; simpl; f_equal; eauto; done); [].
(* use *)
intros h e IH i.
simpl.
f_equal; eauto; [].
#2 destruct h as [c | k].
  simpl; reflexivity.
simpl.
f_equal; [].
unfold sh_var.
remember (lt_dec k i) as b eqn:Heq.
destruct b as [H | H].
  rewrite <- Heq; auto.

  forget (lt_dec (j+k) i) as b.
  destruct b; omega.
Qed.


Lemma sh_head_commute :
  forall i j i' j' h,
    sh_head (i + j + i') j' (sh_head i j h) = sh_head i j (sh_head (i+i') j' h).
Proof.
intros i j i' j' h.
#2 destruct h as [c | k].
  simpl; reflexivity.
simpl.
f_equal; [].
unfold sh_var.
remember (lt_dec k i) as b eqn:Heq.
remember (lt_dec k (i+i')) as b' eqn:Heq'.
#3 destruct b as [H | H]; destruct b' as [H' | H']; try omega.
  rewrite <- Heq; [].
  forget (lt_dec k (i+j+i')) as b; destruct b; omega.

  rewrite <- Heq; [].
  forget (lt_dec (j+k) (i+j+i')) as b; destruct b; omega.

  forget (lt_dec (j+k) (i+j+i')) as b; forget (lt_dec (j'+k) i) as b'.
  destruct b; destruct b'; omega.
Qed.


Lemma sh_commute :
  forall {z} i j i' j' (e : exp z),
    sh (i + j + i') j' (sh i j e) = sh i j (sh (i+i') j' e).
Proof.
intros z i j i' j' e.
revert i i'.
#7 induct e; try (intros; simpl; f_equal; eauto; done).
(* use *)
intros h e IH i i'.
simpl.
f_equal; eauto; [].
apply sh_head_commute; auto; done.

(* pm *)
intros p e H i i'.
simpl.
replace (bound p + (i + j + i')) with ((bound p + i) + j + i') by omega.
replace (bound p + (i + i')) with ((bound p + i) + i') by omega.
f_equal; auto; done.

(* lam *)
intros p e H i i'.
simpl.
replace (bound p + (i + j + i')) with ((bound p + i) + j + i') by omega.
replace (bound p + (i + i')) with ((bound p + i) + i') by omega.
f_equal; auto; done.

(* pi *)
intros p A IH1 B IH2 i i'.
simpl.
replace (bound p + (i + j + i')) with ((bound p + i) + j + i') by omega.
replace (bound p + (i + i')) with ((bound p + i) + i') by omega.
f_equal; auto; done.

(* sigma *)
intros A IH1 B IH2 i i'.
simpl.
f_equal; auto; [].
repl (S (i + j + i')) with (S i + j + i') by omega.
repl (S (i + i')) with (S i + i') by omega.
auto; done.

(* ex *)
intros p A IH1 B IH2 i i'.
simpl.
replace (bound p + (i + j + i')) with ((bound p + i) + j + i') by omega.
replace (bound p + (i + i')) with ((bound p + i) + i') by omega.
f_equal; auto; done.

(* tpi *)
intros p A IH1 B IH2 i i'.
simpl.
replace (bound p + (i + j + i')) with ((bound p + i) + j + i') by omega.
replace (bound p + (i + i')) with ((bound p + i) + i') by omega.
f_equal; auto; done.
Qed.


Lemma sh_sum_left :
  forall {z} i j j' (e : exp z),
    sh (i+j) j' (sh i j e) = sh i (j'+j) e.
Proof.
intros z i j j' e.
replace (i + j) with (i + j + 0) by omega.
rewrite -> (sh_commute i j 0 j'); [].
replace (i+0) with i by omega.
replace (j'+j) with (j+j') by omega.
apply sh_sum; done.
Qed.


Lemma sh_sum_middle :
  forall z i j i' j' (e : exp z),
    i <= i'
    -> i' <= i + j
    -> sh i' j' (sh i j e) = sh i (j+j') e.
Proof.
intros z i j i' j' e Hlower Hupper.
repl j with ((i' - i) + (j - (i' - i))) at 1 by omega.
rewrite <- sh_sum at 1; [].
repl i' with (i + (i' - i)) at 1 by omega.
rewrite -> sh_sum_left; [].
rewrite -> sh_sum; [].
f_equal; omega.
Qed.


(* Cancellation *)

Lemma commute_left :
  (forall z a v p i i' j (e e' : exp z),
     cutpf a v p i e e'
     -> cutpf a (sh i' j v) p i (sh (i + bound p + i') j e) (sh (i + i') j e'))
  /\
  (forall a m s r i j,
     cutmf a m s r
     -> cutmf a (sh i j m) (sh i j s) (sh i j r))
  /\
  (forall z a f i i' j (e e' : exp z),
     rsubf a f i e e'
     -> rsubf a (sh i' j f) i (sh (i+1+i') j e) (sh (i+i') j e'))
  /\
  (forall {z} a i j (e : exp z) p r e',
    lsubf a e p r e'
    -> lsubf a (sh i j e) p (sh (bound p + i) j r) (sh i j e')).
Proof.
#7 exploit
  (substitutionf_ind
     (fun z a v p i e e' =>
        forall i' j,
          cutpf a (sh i' j v) p i (sh (i + bound p + i') j e) (sh (i + i') j e'))
     (fun a m s r =>
        forall i j,
          cutmf a (sh i j m) (sh i j s) (sh i j r))
     (fun z a f i e e' =>
        forall i' j,
          rsubf a (sh i' j f) i (sh (i+1+i') j e) (sh (i+i') j e'))
     (fun z a e p r e' =>
        forall i j,
          lsubf a (sh i j e) p (sh (bound p + i) j r) (sh i j e')))
  as Hind;
  try (intros; simpl; repeat (rewrite -> plus_assoc); eauto using @cutpf, cutmf, @rsubf, @lsubf; done).

(* one *)
intros z a i e i' j.
simpl.
repl (i + 0 + i') with (i + i') by omega.
apply cutpf_one; done.

(* tens *)
intros z a v v' p p' i e e' e'' _ IH1 _ IH2 i' j.
simpl.
repl (i + (bound p + bound p') + i') with (bound p' + i + bound p + i') at 1 by omega.
eapply cutpf_tens; eauto; [].
force_exact (IH2 i' j).
f_equal; f_equal; omega.

(* inst *)
intros a f i h s h' s' Hsubh I IH i' j.
simpl.
apply rsubf_inst; eauto; [].
#3 invert Hsubh; clear Hsubh.
  intros c <- <-.
  apply rsub_const; done.

  intros k Hlt <- <-.
  simpl.
  unfold sh_var.
  forget (lt_dec k (i+1+i')) as b; forget (lt_dec k (i+i')) as b'.
  destruct b as [H | H]; destruct b' as [H' | H']; try omega; [].
  apply rsub_var_lt; auto; done.

  intros k Hgeq <- <-.
  simpl.
  unfold sh_var.
  forget (lt_dec k (i+1+i')) as b; forget (lt_dec (k-1) (i+i')) as b'.
  #2 destruct b; destruct b'; try omega.
    apply rsub_var_gt; auto; done.
    
    repl (j + (k - 1)) with (j + k - 1) by omega.
    apply rsub_var_gt; omega.

(* use/same *)
intros a f i s r s' _ IH1 _ IH2 i' j.
simpl.
unfold sh_var.
forget (lt_dec i (i+1+i')) as b; destruct b as [H | ?]; [| omega].
clear H.
eapply rsubf_reduce; eauto; [].
force_exact (IH2 (i+i') j).
f_equal; [].
repl (i + i') with (0 + i + i') by omega.
rewrite -> sh_commute; [].
reflexivity.

(* sigma *)
intros a f i A B A' B' _ IH1 _ IH2 i' j.
simpl.
eapply rsubf_sigma; eauto; [].
force_exact (IH2 i' j).
reflexivity.

(* lsub pm *)
intros a p1 r1 p2 r2 r _ IH i j.
simpl.
eapply lsubf_pm; eauto; [].
force_exact (IH (bound p1 + i) j).
f_equal; [].
rewrite -> plus_assoc; [].
apply sh_commute; done.

(* wrapup *)
destruct_all Hind.
do2 3 split; auto; done.
Qed.


Lemma cutpf_sh_commute_left :
  forall {z} a v p i i' j (e e' : exp z),
    cutpf a v p i e e'
    -> cutpf a (sh i' j v) p i (sh (i + bound p + i') j e) (sh (i + i') j e').
Proof.
exact (commute_left andel).
Qed.


Lemma cutmf_sh_commute :
  forall a m s r i j,
    cutmf a m s r
    -> cutmf a (sh i j m) (sh i j s) (sh i j r).
Proof.
exact (commute_left anderl).
Qed.


Lemma rsubf_sh_commute_left :
  forall {z} a f i i' j (e e' : exp z),
    rsubf a f i e e'
    -> rsubf a (sh i' j f) i (sh (i+1+i') j e) (sh (i+i') j e').
Proof.
exact (commute_left anderrl).
Qed.


Lemma sh_lsubf :
  forall {z} a i j (E : exp z) p r E',
    lsubf a E p r E'
    -> lsubf a (sh i j E) p (sh (bound p + i) j r) (sh i j E').
Proof.
exact (commute_left anderrr).
Qed.


Lemma sh_lsub :
  forall {z} i j (E : exp z) p r E',
    lsub E p r E'
    -> lsub (sh i j E) p (sh (bound p + i) j r) (sh i j E').
Proof.
intros z i j E p r E' Hsub.
so (fuel_lsub _#4 Hsub) as (a, Hsub').
eauto using sh_lsubf, unfuel_lsub.
Qed.


Lemma cutp_sh_commute_left :
  forall z v p i i' j (e e' : exp z),
    cutp v p i e e'
    -> cutp (sh i' j v) p i (sh (i + bound p + i') j e) (sh (i + i') j e').
Proof.
intros z v p i i' j e e' Hcut.
so (fuel_cutp _#6 Hcut) as (a & Hcutf).
so (cutpf_sh_commute_left _#4 i' j _ _ Hcutf) as Hcutf'.
eapply unfuel_cutp; eauto.
Qed.


Lemma cutm_sh_commute :
  forall m s r i j,
    cutm m s r
    -> cutm (sh i j m) (sh i j s) (sh i j r).
Proof.
intros m s r i j Hcut.
so (fuel_cutm _#3 Hcut) as (a & Hcutf).
so (cutmf_sh_commute _#4 i j Hcutf) as Hcutf'.
eapply unfuel_cutm; eauto.
Qed.


Lemma rsub_sh_commute_left :
  forall {z} f i i' j (e e' : exp z),
    rsub f i e e'
    -> rsub (sh i' j f) i (sh (i+1+i') j e) (sh (i+i') j e').
Proof.
intros z f i i' j e e' Hsub.
so (fuel_rsub _#4 Hsub) as (a, Hsubf).
apply (unfuel_rsub a); [].
eapply rsubf_sh_commute_left; eauto.
Qed.


Lemma rsubf_sh_commute_right :
  forall {z} a f i i' j (e e' : exp z),
    i' <= i
    -> rsubf a f i e e'
    -> rsubf a f (i+j) (sh i' j e) (sh i' j e').
Proof.
intros z a f i i' j e e' Hleq H.
revert i' j Hleq.
#8 induct H; try (intros; simpl; eauto using @rsubf; done).

(* inst *)
intros a f i h s h' s' Hsubh _ IH i' j Hleq.
simpl.
apply rsubf_inst; eauto; [].
#3 invert Hsubh; clear Hsubh.
  intros c <- <-.
  apply rsub_const; done.

  intros k Hlt <- <-.
  apply rsub_var_lt; [].
  unfold sh_var.
  forget (lt_dec k i') as b; destruct b; omega.

  intros k Hgeq <- <-.
  simpl.
  unfold sh_var.
  forget (lt_dec k i') as b; forget (lt_dec (k - 1) i') as b'.
  destruct b as [H | H]; destruct b' as [H' | H']; try omega; [].
  replace (j + (k - 1)) with ((j + k) - 1) by omega.
  apply rsub_var_gt; omega.

(* reduce *)
intros a f i s r s' _ IH Hcut i' j Hleq.
simpl.
unfold sh_var.
forget (lt_dec i i') as b; destruct b as [? | Hnlt]; [omega |].
repl (j + i) with (i + j) by omega.
eapply rsubf_reduce; eauto; [].
#2 repl (sh 0 (i+j) f) with (sh i' j (sh 0 i f)).
  apply sh_sum_middle; omega.
apply cutmf_sh_commute; auto; done.

(* pm *)
intros a f i p r r' _ IH i' j Hleq.
simpl.
apply rsubf_pm; [].
rewrite -> plus_assoc; [].
apply IH; omega.

(* lam *)
intros a f i p n n' _ IH i' j Hleq.
simpl.
apply rsubf_lam; [].
rewrite -> plus_assoc; [].
apply IH; omega.

(* pi *)
intros a f i p A B A' B' _ IH1 _ IH2 i' j Hleq.
simpl.
apply rsubf_pi; auto; [].
rewrite -> plus_assoc; [].
apply IH2; omega.

(* pi *)
intros a f i A B A' B' _ IH1 _ IH2 i' j Hleq.
simpl.
apply rsubf_sigma; auto; [].
rewrite -> plus_assoc; [].
apply IH2; omega.

(* ex *)
intros a f i p A B A' B' _ IH1 _ IH2 i' j Hleq.
simpl.
apply rsubf_ex; auto; [].
rewrite -> plus_assoc; [].
apply IH2; omega.

(* tpi *)
intros a f i p A B A' B' _ IH1 _ IH2 i' j Hleq.
simpl.
apply rsubf_tpi; auto; [].
rewrite -> plus_assoc; [].
apply IH2; omega.
Qed.


Lemma rsub_sh_commute_right :
  forall {z} f i i' j (e e' : exp z),
    i' <= i
    -> rsub f i e e'
    -> rsub f (i+j) (sh i' j e) (sh i' j e').
Proof.
intros z f i i' j e e' Hleq Hsub.
so (fuel_rsub _#4 Hsub) as (a, Hsub').
eauto using rsubf_sh_commute_right, unfuel_rsub.
Qed.


Lemma cutp_sh_commute_right :
  forall {z} v p i i' j (e e' : exp z),
    i' <= i
    -> cutp v p i e e'
    -> cutp v p (i+j) (sh i' j e) (sh i' j e').
Proof.
intros z v p i i' j e e' Hleq Hcut.
revert i' Hleq.
induct Hcut.

(* down *)
intros z m i e e' Hsub i' Hleq.
apply cutp_down; [].
eapply rsub_sh_commute_right; eauto; done.

(* one *)
intros z i e i' Hleq.
apply cutp_one; done.

(* tens *)
intros z v v' p p' i e e' e'' _ IH1 _ IH2 i' Hleq.
eapply cutp_tens; eauto; [].
repl (bound p' + (i + j)) with (bound p' + i + j) by omega.
apply IH1; omega.
Qed.


Lemma rsub_sh_commute_right_0 :
  forall {z} f i j (e e' : exp z),
    rsub f i e e'
    -> rsub f (i+j) (sh 0 j e) (sh 0 j e').
Proof.
intros z f i j e e' H.
eapply rsub_sh_commute_right; eauto; [].
omega.
Qed.


Lemma rsub_head_sh_head_cancel :
  forall i i' j h,
    i >= i'
    -> i < i' + j
    -> rsub_head i (sh_head i' j h) (sh_head i' (j - 1) h).
Proof.
intros i i' j h Hlower Hupper.
destruct h as [c | k].
  simpl.
  apply rsub_const; done.
unfold sh_head, sh_var.
forget (lt_dec k i') as b; destruct b as [Hlt | Hnlt].
  apply rsub_var_lt; omega.

  repl (j - 1 + k) with (j + k - 1) by omega.
  apply rsub_var_gt; omega.
Qed.


Lemma rsubf_sh_cancel :
  forall {z} a f i i' j (e : exp z),
    i >= i'
    -> i < i' + j
    -> rsubf a f i (sh i' j e) (sh i' (j-1) e).
Proof.
intros z a f i i' j e Hlower Hupper.
revert i i' Hlower Hupper.
#7 induct e; try (intros; simpl; eauto using @rsubf; done).

(* use *)
intros h s IH i i' Hlower Hupper.
simpl.
apply rsubf_inst; auto; [].
destruct h as [c | k].
  simpl.
  apply rsub_const; done.
unfold sh_head, sh_var.
forget (lt_dec k i') as b; destruct b as [Hlt | Hnlt].
  apply rsub_var_lt; omega.
  repl (j - 1 + k) with (j + k - 1) by omega.
  apply rsub_var_gt; omega.

(* pm *)
intros p e IH i i' Hlower Hupper.
simpl.
apply rsubf_pm; [].
apply IH; omega.

(* lam *)
intros p e IH i i' Hlower Hupper.
simpl.
apply rsubf_lam; [].
apply IH; omega.

(* pi *)
intros p A IH1 B IH2 i i' Hlower Hupper.
simpl.
apply rsubf_pi; eauto; [].
apply IH2; omega.

(* sigma *)
intros A IH1 B IH2 i i' Hlower Hupper.
simpl.
apply rsubf_sigma; eauto; [].
apply IH2; omega.

(* ex *)
intros p A IH1 B IH2 i i' Hlower Hupper.
simpl.
apply rsubf_ex; eauto; [].
apply IH2; omega.

(* tpi *)
intros p A IH1 B IH2 i i' Hlower Hupper.
simpl.
apply rsubf_tpi; eauto; [].
apply IH2; omega.
Qed.


Lemma rsub_sh_cancel :
  forall {z} f i i' j (e : exp z),
    i >= i'
    -> i < i' + j
    -> rsub f i (sh i' j e) (sh i' (j-1) e).
Proof.
intros; eapply unfuel_rsub; eapply (rsubf_sh_cancel 0); eauto.
Qed.  


Lemma rsub_sh_cancel_all :
  forall {z} f i (e : exp z),
    rsub f i (sh i 1 e) e.
Proof.
intros z f i e.
exploit (rsub_sh_cancel f i i 1 e) as H; try omega; [].
repl (1 - 1) with 0 in H by omega.
rewrite -> sh_zero in H; [].
exact H.
Qed.


Lemma cutpf_sh_cancel :
  forall {z} a v p i i' j (e : exp z),
    i >= i'
    -> i + bound p <= i' + j
    -> patmatch v p
    -> cutpf a v p i (sh i' j e) (sh i' (j - bound p) e).
Proof.
intros z a v p i i' j e Hlower Hupper Hmatch.
revert i i' j e Hlower Hupper.
induct Hmatch.

(* down *)
intros m i i' j e Hlower Hupper.
simpl in Hupper |- *.
apply cutpf_down; [].
apply rsubf_sh_cancel; eauto; omega.

(* one *)
intros i i' j e Hlower Hupper.
simpl.
repl (j - 0) with j by omega.
apply cutpf_one; done.

(* tens *)
intros v w p q _ IH1 _ IH2 i i' j e Hlower Hupper.
simpl in Hupper.
simpl.
exploit (IH1 (bound q + i) i' j e) as Hcut1; try omega; [].
exploit (IH2 i i' (j - bound p) e) as Hcut2; try omega; [].
eapply cutpf_tens; eauto; [].
repl (j - (bound p + bound q)) with (j - bound p - bound q) by omega.
exact Hcut2.
Qed.


Lemma cutp_sh_cancel :
  forall {z} v p i i' j (e : exp z),
    i >= i'
    -> i + bound p <= i' + j
    -> patmatch v p
    -> cutp v p i (sh i' j e) (sh i' (j - bound p) e).
Proof.
intros z v p i i' j e Hlower Hupper Hmatch.
so (cutpf_sh_cancel 0 v p i i' j e Hlower Hupper Hmatch) as Hcutf.
eapply unfuel_cutp; eauto.
Qed.


Lemma cutp_sh_cancel_all :
  forall {z} v p i (e : exp z),
    patmatch v p
    -> cutp v p i (sh i (bound p) e) e.
Proof.
intros z v p i e Hmatch.
rewrite <- (sh_zero i e) at 2.
repl 0 with (bound p - bound p) by omega.
apply cutp_sh_cancel; auto; omega.
Qed.


(* Substitution compostition & commutativity *)

Lemma commute_rsub_head :
  forall i1 i2 h h1 h2,
    i2 < i1
    -> rsub_head i1 h h1
    -> rsub_head i2 h h2
    -> exists h',
         rsub_head i2 h1 h'
         /\ rsub_head (i1-1) h2 h'.
Proof.
intros i1 i2 h h1 h2 Hleq Hsub1 Hsub2.
revert Hleq Hsub2.
induct Hsub1.

(* const *)
intros i1 c Hleq Hsub2.
invert Hsub2; [].
intros <-.
exists (const c).
split; apply rsub_const; done.

(* lt *)
intros i1 k Hlt1 Hleq Hsub2.
#2 invert Hsub2; clear Hsub2.
  (* lt *)
  intros Hlt2 <-.
  exists (var k).
  split; apply rsub_var_lt; omega.

  (* gt *)
  intros Hgt <-.
  exists (var (k-1)).
  split.
    apply rsub_var_gt; omega.

    apply rsub_var_lt; omega.

(* gt *)
intros i1 k Hgt1 Hleq Hsub2.
#2 invert Hsub2; clear Hsub2.
  (* lt *)
  intros Hlt <-.
  exfalso.
  omega.

  (* gt *)
  intros Hgt2 <-.
  exists (var (k-1-1)).
  split.
    apply rsub_var_gt; omega.

    apply rsub_var_gt; omega.
Qed.



Definition commute_cutm_rsub_prop (N : nat) :=
  forall a b m s r f i m' s',
    a + b = N
    -> cutmf a m s r
    -> rsubf b f i m m'
    -> rsubf b f i s s'
    -> exists r',
         rsubf b f i r r'
         /\ cutmf a m' s' r'.

Definition commute_cutp_rsub_prop (N : nat) :=
  forall z a1 a2 v p i1 f i2 (E E1 E2 : exp z) v',
    a1 + a2 = N
    -> i2 + bound p <= i1
    -> rsubf a1 f i1 E E1
    -> cutpf a2 v p i2 E E2
    -> rsubf a1 f (i1 - i2 - bound p) v v'
    -> exists E',
         cutpf a2 v' p i2 E1 E'
         /\ rsubf a1 f (i1 - bound p) E2 E'.

Definition commute_rsub_prop (N : nat) :=
  forall z a1 a2 f1 i1 f2 i2 (E E1 E2 : exp z) f2',
    a1 + a2 = N
    -> i2 < i1
    -> rsubf a1 f1 i1 E E1
    -> rsubf a2 f2 i2 E E2
    -> rsubf a1 f1 (i1 - i2 - 1) f2 f2'
    -> exists E',
         rsubf a2 f2' i2 E1 E'
         /\ rsubf a1 f1 (i1-1) E2 E'.

Definition commute_lsub_rsub_prop (N : nat) :=
  forall z a b (E1 : exp z) p r E2 f i E1' r',
    a + b = N
    -> lsubf a E1 p r E2
    -> rsubf b f i E1 E1'
    -> rsubf b f (bound p + i) r r'
    -> exists E2',
         rsubf b f i E2 E2'
         /\ lsubf a E1' p r' E2'.

Definition commute_lsub_cutp_prop (N : nat) :=
  forall z a b (E1 : exp z) p r E2 v q i E1' r',
    a + b = N
    -> lsubf a E1 p r E2
    -> cutpf b v q i E1 E1'
    -> cutpf b v q (bound p + i) r r'
    -> exists E2',
         cutpf b v q i E2 E2'
         /\ lsubf a E1' p r' E2'.

Definition commute_cutm_lsub_prop (N : nat) :=
  forall a b m s r1 p r2 s',
    a + b = N
    -> cutmf a m s r1
    -> lsubf b s p r2 s'
    -> exists r,
         lsubf b r1 p r2 r
         /\ cutmf a m s' r.

Definition commute_lsub_prop (N : nat) :=
  forall z a1 a2 (E1 : exp z) p1 r2 p2 r3 E12 r23,
    a1 + a2 = N
    -> lsubf a1 E1 p1 r2 E12
    -> lsubf a2 r2 p2 (sh (bound p2) (bound p1) r3) r23
    -> exists E,
         lsubf a1 E1 p1 r23 E
         /\ lsubf a2 E12 p2 r3 E.


Lemma commutativity :
  forall N,
    commute_rsub_prop N
    /\ commute_cutp_rsub_prop N
    /\ commute_lsub_rsub_prop N
    /\ commute_lsub_cutp_prop N
    /\ commute_lsub_prop N
    /\ commute_cutm_rsub_prop N
    /\ commute_cutm_lsub_prop N.
Proof.
intros N.
wfinduct N using lt_wf.
intros N IH.
so (fun N' Hlt => IH N' Hlt anderrrrrl) as IHmr.
so (fun N' Hlt => IH N' Hlt anderrrrrr) as IHml.
clear IH.
assert (commute_rsub_prop N) as Hr.
  (* commute_rsub *)
  clear IHml.
  unfold commute_rsub_prop.
  intros z a1 a2 f1 i1 f2 i2 E E1 E2 f2' Hsum Hleq Hsub1 Hsub2 Hsubf.
  revert f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  induct Hsub1.
  (* ret *)
  intros a1 f i1 v v1 Hsubv1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf2.
  invert Hsub2; clear Hsub2; [].
  intros v2 Hsubv2 <-.
  so (IH _#4 Hsum Hleq Hsubv2 Hsubf2) as (v' & Hsubv1' & Hsubv2').
  exists (ret v').
  split; eauto using @rsubf; done.

  (* inst *)
  intros a1 f1 i1 h s h1 s1 Hsubh1 Hsubs1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf2.
  #2 invert Hsub2; clear Hsub2.
    (* inst *)
    intros h2 s2 Hsubh2 Hsubs2 <-.
    so (IH _#4 Hsum Hleq Hsubs2 Hsubf2) as (s' & Hsubs1' & Hsubs2').
    so (commute_rsub_head _#5 Hleq Hsubh1 Hsubh2) as (h' & Hsubh1' & Hsubh2').
    exists (use h' s').
    split; eauto using @rsubf; done.

    (* reduce *)
    intros a2' s2 Hsubs2 Hcut <- <-.
    assert (h1 = var i2).
      #2 invert Hsubh1.
        intros _ <-.
        reflexivity.
    
        intros ?.
        exfalso.
        omega.
    subst h1; clear Hsubh1.
    so (IH _#4 Hsum Hleq Hsubs2 Hsubf2) as (s' & Hsubs1' & Hsubs2'); clear IH.
    exploit (rsubf_sh_commute_right a1 f1 (i1-i2-1) 0 i2 f2 f2') as Hsubf2'; [omega | exact Hsubf2 |].
    replace (i1 - i2 - 1 + i2) with (i1-1) in Hsubf2' by omega.
    assert (a2' + a1 < N) as Hlt by omega.
    unfold commute_cutm_rsub_prop in IHmr.
    so (IHmr _ Hlt _#9 (eq_refl _) Hcut Hsubf2' Hsubs2') as (r' & Hsub & Hcut').
    clear IHmr.
    exists r'.
    split; auto; [].
    eapply rsubf_reduce; eauto; done.

  (* reduce *)
  intros a1 f1 i1 s r s1 Hsubs1 IH1 Hcut f2 i2 E2 f2' Hsum Hlt Hsub2 Hsubf2.
  #2 invert Hsub2; clear Hsub2.
    (* use/diff *)
    intros h2 s2 Hsubh2 Hsubs2 <-.
    assert (h2 = var (i1-1)).
      #2 invert Hsubh2.
        intros ? _.
        exfalso.
        omega.
  
        intros _ <-.
        reflexivity.
    subst h2; clear Hsubh2.
    so (IH1 _#4 Hsum Hlt Hsubs2 Hsubf2) as (s' & Hsubs1' & Hsubs2'); clear IH1.
    exploit (rsubf_sh_cancel a2 f2' i2 0 i1 f1) as Hsubm; try omega; [].
    assert (a1 + a2 < N) as Hltsize by omega.
    unfold commute_cutm_rsub_prop in IHmr.
    so (IHmr _ Hltsize _#9 (eq_refl _) Hcut Hsubm Hsubs1') as (r' & Hsub & Hcut').
    clear IHmr.
    exists r'.
    split; auto; [].
    eapply rsubf_reduce; eauto; done.
  
    (* inst *)
    intros.
    exfalso.
    omega.
  
  (* thunk *)
  intros a1 f1 i1 m m1 Hsubm1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros m2 Hsubm2 <-.
  so (IH _#4 Hsum Hleq Hsubm2 Hsubf) as (m' & Hsubm1' & Hsubm2').
  exists (thunk m').
  split; eauto using @rsubf; done.
  
  (* unitp *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros <-.
  exists unitp.
  split; eauto using @rsubf; done.

  (* pairp *)
  intros a1 f i v w v1 w1 Hsubv1 IH1 Hsubw1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros v2 w2 Hsubv2 Hsubw2 <-.
  so (IH1 _#4 Hsum Hleq Hsubv2 Hsubf) as (v' & Hsubv1' & Hsubv2').
  so (IH2 _#4 Hsum Hleq Hsubw2 Hsubf) as (w' & Hsubw1' & Hsubw2').
  clear IH1 IH2.
  exists (pairp v' w').
  split; eauto using @rsubf; done.

  (* nl *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; [].
  intros <-.
  exists nl.
  split; eauto using @rsubf; done.

  (* pm *)
  intros a1 f1 i1 p r r1 Hsubr1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros r2 Hsubr2 <-.
  #3 sor (refine (IH _#4 Hsum _ Hsubr2 _)) as H.
    omega.
  
    replace (bound p + i1 - (bound p + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (r' & Hsubr1' & Hsubr2').
  exists (pm p r').
  split; eauto using @rsubf; [].
  apply rsubf_pm; [].
  replace (bound p + (i1 - 1)) with (bound p + i1 - 1) by omega.
  exact Hsubr2'.

  (* app *)
  intros a1 f i v s v1 s1 Hsubv1 IH1 Hsubs1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros v2 s2 Hsubv2 Hsubs2 <-.
  so (IH1 _#4 Hsum Hleq Hsubv2 Hsubf) as (v' & Hsubv1' & Hsubv2').
  so (IH2 _#4 Hsum Hleq Hsubs2 Hsubf) as (s' & Hsubs1' & Hsubs2').
  clear IH1 IH2.
  exists (app v' s').
  split; eauto using @rsubf; done.

  (* pi1 *)
  intros a1 f i s s1 Hsubs1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros s2 Hsubs2 <-.
  so (IH _#4 Hsum Hleq Hsubs2 Hsubf) as (s' & Hsubs1' & Hsubs2').
  exists (pi1 s').
  split; eauto using @rsubf; done.

  (* pi2 *)
  intros a1 f i s s1 Hsubs1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros s2 Hsubs2 <-.
  so (IH _#4 Hsum Hleq Hsubs2 Hsubf) as (s' & Hsubs1' & Hsubs2').
  exists (pi2 s').
  split; eauto using @rsubf; done.

  (* etam *)
  intros a1 f i r r1 Hsubr1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros r2 Hsubr2 <-.
  so (IH _#4 Hsum Hleq Hsubr2 Hsubf) as (r' & Hsubr1' & Hsubr2').
  exists (etam r').
  split; eauto using @rsubf; done.

  (* shiftp *)
  intros a1 f i r r1 Hsubr1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros r2 Hsubr2 <-.
  so (IH _#4 Hsum Hleq Hsubr2 Hsubf) as (r' & Hsubr1' & Hsubr2').
  exists (shiftp r').
  split; eauto using @rsubf; done.
  
  (* lam *)
  intros a1 f1 i1 p m m1 Hsubm1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros m2 Hsubm2 <-.
  #3 sor (refine (IH _#4 Hsum _ Hsubm2 _)) as H.
    omega.
  
    replace (bound p + i1 - (bound p + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (m' & Hsubm1' & Hsubm2').
  exists (lam p m').
  split; eauto using @rsubf; [].
  apply rsubf_lam; [].
  replace (bound p + (i1 - 1)) with (bound p + i1 - 1) by omega.
  exact Hsubm2'.
  
  (* pairm *)
  intros a1 f i m n m1 n1 Hsubm1 IH1 Hsubn1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros m2 n2 Hsubm2 Hsubn2 <-.
  so (IH1 _#4 Hsum Hleq Hsubm2 Hsubf) as (m' & Hsubm1' & Hsubm2').
  so (IH2 _#4 Hsum Hleq Hsubn2 Hsubf) as (n' & Hsubn1' & Hsubn2').
  clear IH1 IH2.
  exists (pairm m' n').
  split; eauto using @rsubf; done.
  
  (* tuse *)
  intros a1 f1 i1 c s s1 Hsubs1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros s2 Hsubs2 <-.
  so (IH _#4 Hsum Hleq Hsubs2 Hsubf) as (s' & Hsubs1' & Hsubs2').
  exists (tuse c s').
  split; eauto using @rsubf; done.
  
  (* up *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (up A').
  split; eauto using @rsubf; done.
  
  (* pi *)
  intros a1 f1 i1 p A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  #3 sor (refine (IH2 _#4 Hsum _ HsubB2 _)) as H.
    omega.
  
    replace (bound p + i1 - (bound p + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (B' & HsubB1' & HsubB2').
  exists (pi p A' B').
  split; eauto using @rsubf; [].
  apply rsubf_pi; auto; [].
  replace (bound p + (i1 - 1)) with (bound p + i1 - 1) by omega.
  exact HsubB2'.
  
  (* wth *)
  intros a1 f i A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  so (IH2 _#4 Hsum Hleq HsubB2 Hsubf) as (B' & HsubB1' & HsubB2').
  clear IH1 IH2.
  exists (wth A' B').
  split; eauto using @rsubf; done.

  (* sigma *)
  intros a1 f1 i1 A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  #3 sor (refine (IH2 _#4 Hsum _ HsubB2 _)) as H.
    omega.
  
    replace (1 + i1 - (1 + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (B' & HsubB1' & HsubB2').
  exists (sigma A' B').
  split; eauto using @rsubf; [].
  apply rsubf_sigma; auto; [].
  replace (1 + (i1 - 1)) with (1 + i1 - 1) by omega.
  exact HsubB2'.
  
  (* down *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (down A').
  split; eauto using @rsubf; done.
  
  (* top *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; [].
  intros <-.
  exists one.
  split; eauto using @rsubf; done.

  (* tens *)
  intros a1 f i A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  so (IH2 _#4 Hsum Hleq HsubB2 Hsubf) as (B' & HsubB1' & HsubB2').
  clear IH1 IH2.
  exists (tens A' B').
  split; eauto using @rsubf; done.

  (* ex *)
  intros a1 f1 i1 p A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  #3 sor (refine (IH2 _#4 Hsum _ HsubB2 _)) as H.
    omega.
  
    replace (bound p + i1 - (bound p + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (B' & HsubB1' & HsubB2').
  exists (ex p A' B').
  split; eauto using @rsubf; [].
  apply rsubf_ex; auto; [].
  replace (bound p + (i1 - 1)) with (bound p + i1 - 1) by omega.
  exact HsubB2'.
  
  (* tnl *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; [].
  intros <-.
  exists tnl.
  split; eauto using @rsubf; done.

  (* tapp *)
  intros a1 f i v s v1 s1 Hsubv1 IH1 Hsubs1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros v2 s2 Hsubv2 Hsubs2 <-.
  so (IH1 _#4 Hsum Hleq Hsubv2 Hsubf) as (v' & Hsubv1' & Hsubv2').
  so (IH2 _#4 Hsum Hleq Hsubs2 Hsubf) as (s' & Hsubs1' & Hsubs2').
  clear IH1 IH2.
  exists (tapp v' s').
  split; eauto using @rsubf; done.

  (* ty *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; [].
  intros <-.
  exists ty.
  split; eauto using @rsubf; done.

  (* tpi *)
  intros a1 f1 i1 p A K A1 K1 HsubA1 IH1 HsubK1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 K2 HsubA2 HsubK2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  #3 sor (refine (IH2 _#4 Hsum _ HsubK2 _)) as H.
    omega.
  
    replace (bound p + i1 - (bound p + i2) - 1) with (i1 - i2 - 1) by omega.
    exact Hsubf.
  destruct H as (K' & HsubK1' & HsubK2').
  exists (tpi p A' K').
  split; eauto using @rsubf; [].
  apply rsubf_tpi; auto; [].
  replace (bound p + (i1 - 1)) with (bound p + i1 - 1) by omega.
  exact HsubK2'.
  
  (* trp *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (trp A').
  split; eauto using @rsubf; done.

  (* susp *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (susp A').
  split; eauto using @rsubf; done.

  (* judr *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (judr A').
  split; eauto using @rsubf; done.

  (* judv *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (judv A').
  split; eauto using @rsubf; done.

  (* juds *)
  intros a1 f i A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  so (IH2 _#4 Hsum Hleq HsubB2 Hsubf) as (B' & HsubB1' & HsubB2').
  clear IH1 IH2.
  exists (juds A' B').
  split; eauto using @rsubf; done.

  (* judt *)
  intros a1 f i A A1 HsubA1 IH f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.  
  invert Hsub2; clear Hsub2; [].
  intros A2 HsubA2 <-.
  so (IH _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  exists (judt A').
  split; eauto using @rsubf; done.

  (* judts *)
  intros a1 f i A B A1 B1 HsubA1 IH1 HsubB1 IH2 f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invert Hsub2; clear Hsub2; [].
  intros A2 B2 HsubA2 HsubB2 <-.
  so (IH1 _#4 Hsum Hleq HsubA2 Hsubf) as (A' & HsubA1' & HsubA2').
  so (IH2 _#4 Hsum Hleq HsubB2 Hsubf) as (B' & HsubB1' & HsubB2').
  clear IH1 IH2.
  exists (judts A' B').
  split; eauto using @rsubf; done.

  (* judnt *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invertc Hsub2; [].
  intros <-.
  exists judnt.
  split; eauto using @rsubf; done.

  (* judpt *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invertc Hsub2; [].
  intros <-.
  exists judpt.
  split; eauto using @rsubf; done.

  (* judk *)
  intros a1 f i f2 i2 E2 f2' Hsum Hleq Hsub2 Hsubf.
  invertc Hsub2; [].
  intros <-.
  exists judk.
  split; eauto using @rsubf; done.
split; [exact Hr |].
assert (commute_cutp_rsub_prop N) as Hpr.
  clear IHmr IHml.
  unfold commute_cutp_rsub_prop.
  intros z a b v p i1 f i2 e e1 e2 v' Hsum Hleq Hsub1 Hcutp Hsubv.
  revert a i1 f e1 v' Hsum Hleq Hsub1 Hsubv.
  induct Hcutp.
    (* down *)
    intros z b m i2 e e2 Hsub2 a i1 f e1 v' Hsum Hleq Hsub1 Hsubv.
    simpl in Hleq, Hsubv |- *.
    invertc Hsubv; [].
    intros m' Hsubm <-.
    unfold commute_rsub_prop in Hr.
    assert (i2 < i1) as Hlt by omega.
    so (Hr _#11 Hsum Hlt Hsub1 Hsub2 Hsubm) as (e' & Hsub1' & Hsub2').
    exists e'.
    split; eauto using @cutpf; done.

    (* one *)
    intros z b i2 e a i1 f e1 v' Hsum Hleq Hsub1 Hsubv.
    simpl in Hleq, Hsubv |- *.
    invertc Hsubv; [].
    intros <-.
    exists e1.
    repl (i1 - 0) with i1 by omega.
    split; eauto using @cutpf; done.

    (* tens *)
    intros z b v w p q i e e1 e1' Hcut1 IH1 Hcut2 IH2 a i1 f e2 vv Hsum Hleq Hsub2 Hsubv.
    simpl in Hleq, Hsubv |- *.
    invertc Hsubv; [].
    intros v' w' Hsubv Hsubw <-.
    repl (i1 - i - (bound p + bound q)) with (i1 - (bound q + i) - bound p) in Hsubv by omega.
    repl (i1 - i - (bound p + bound q)) with (i1 - bound p - i - bound q) in Hsubw by omega.
    sor (refine (IH1 _#5 Hsum _ Hsub2 Hsubv)) as (e' & Hcut1' & Hsub2'); [omega |].
    sor (refine (IH2 _#5 Hsum _ Hsub2' Hsubw)) as (e'' & Hcut2' & Hsub2''); [omega |].
    exists e''.
    repl (i1 - (bound p + bound q)) with (i1 - bound p - bound q) by omega.
    split; eauto using @cutpf; done.
split; [exact Hpr |].
assert (commute_lsub_rsub_prop N) as Hlr.
  (* commute_lsub_rsub *)
  clear IHmr.
  unfold commute_lsub_rsub_prop.
  intros z a b E1 p r E2 f i E1' r' Hsum Hlsubf HsubE1 Hsubr.
  revert f i E1' r' Hsum HsubE1 Hsubr.
  induct Hlsubf.
  
  (* ret *)
  intros a v p r r2 Hcutp f1 i1 E1' r1 Hsum HsubE1 Hsubr1.
  invert HsubE1; clear HsubE1; [].
  intros v' Hsubv <-.
  unfold commute_cutp_rsub_prop in Hpr.
  repl i1 with (bound p + i1 - 0 - bound p) in Hsubv by omega.
  sor (refine (Hpr _#12 _ _ Hsubr1 Hcutp Hsubv)) as H; [omega | omega |].
  destruct H as (r' & Hcutp' & Hsubr2).
  exists r'.
  split.
    repl (bound p + i1 - bound p) with i1 in Hsubr2 by omega.
    exact Hsubr2.

    apply lsubf_ret; auto; done.
  
  (* use *)
  intros a h s p r s1 Hlsubf IH f i r1 r2 Hsum HsubE1 Hsubr2.
  #2 invert HsubE1; clear HsubE1.
    (* inst *)
    intros h' s2 Hsubh Hsubs2 <-.
    so (IH _#4 Hsum Hsubs2 Hsubr2) as (s' & Hsubs1' & Hlsubf').
    exists (use h' s').
    split; eauto using @rsubf, @lsubf; done.
  
    (* reduce *)
    intros b' s2 Hsubs2 Hcutmf <- <-.
    so (IH _#4 Hsum Hsubs2 Hsubr2) as (s' & Hsubs1 & Hlsubf').
    assert (b' + a < N) as Hlt by omega.
    unfold commute_cutm_lsub_prop in IHml.
    so (IHml _ Hlt _#8 (eq_refl _) Hcutmf Hlsubf') as (r' & Hsub'' & Hcutmf').
    clear IHml.
    exists r'.
    split; eauto using @rsubf, @lsubf; done.
  
  (* pm *)
  intros a p1 r1 p2 r2 r Hlsubf IH f i E1' r2' Hsum HsubE1 Hsubr2.
  invert HsubE1; clear HsubE1; [].
  intros r1' Hsubr1 <-.
  #2 sor (refine (rsubf_sh_commute_right _ _ _ (bound p2) (bound p1) _ _ _ Hsubr2)) as Hsubr2'.
    omega.
  replace (bound p2 + i + bound p1) with (bound p2 + (bound p1 + i)) in Hsubr2' by omega.
  so (IH _#4 Hsum Hsubr1 Hsubr2') as (r' & Hsub & Hlsubf').
  exists (pm p1 r').
  split; eauto using @rsubf, @lsubf; done.
  
  (* app *)
  intros a v s1 p r s2 Hlsubf IH f i E1' r' Hsum HsubE1 Hsubr.
  invert HsubE1; clear HsubE1; [].
  intros v' s1' Hsubv Hsubs1 <-.
  so (IH _#4 Hsum Hsubs1 Hsubr) as (s2' & Hsubs2 & Hlsubf').
  exists (app v' s2').
  split; eauto using @lsubf, @rsubf; done.
  
  (* pi1 *)
  intros a s1 p r s2 Hlsubf IH f i E1' r' Hsum HsubE1 Hsubr.
  invert HsubE1; clear HsubE1; [].
  intros s1' Hsubs1 <-.
  so (IH _#4 Hsum Hsubs1 Hsubr) as (s2' & Hsubs2 & Hlsubf').
  exists (pi1 s2').
  split; eauto using @lsubf, @rsubf; done.
  
  (* pi2 *)
  intros a s1 p r s2 Hlsubf IH f i E1' r' Hsum HsubE1 Hsubr.
  invert HsubE1; clear HsubE1; [].
  intros s1' Hsubs1 <-.
  so (IH _#4 Hsum Hsubs1 Hsubr) as (s2' & Hsubs2 & Hlsubf').
  exists (pi2 s2').
  split; eauto using @lsubf, @rsubf; done.
split; [exact Hlr |].
assert (commute_lsub_cutp_prop N) as Hlp.
  clear IHmr IHml Hr Hpr.
  unfold commute_lsub_cutp_prop.
  intros z a b e p r e1 v q i e2 r' Hsum Hlsub Hcutp Hcutpr.
  revert a p r e1 r' Hsum Hlsub Hcutpr.
  induct Hcutp.
    (* down *)
    intros z b m i e e2 Hrsub a p r e1 r' Hsum Hlsub Hcutpr.
    invertc Hcutpr; [].
    intros Hsubr.
    unfold commute_lsub_rsub_prop in Hlr.
    so (Hlr _#11 Hsum Hlsub Hrsub Hsubr) as (e' & Hcutp & Hlsub').
    exists e'.
    split; eauto using @cutpf; done.

    (* one *)
    intros z b i e a p r e1 r' Hsum Hlsub Hcutpr.
    invertc Hcutpr; [].
    intros <-.
    exists e1; split; eauto using @cutpf; done.

    (* tens *)
    intros z b v w p1 p2 i e e2 e2' Hcut1 IH1 Hcut2 IH2 a p r e1 r2 Hsum Hlsub Hcutpr.
    invertc Hcutpr; [].
    intros r1 Hcutpr1 Hcutpr2.
    repl (bound p2 + (bound p + i)) with (bound p + (bound p2 + i)) in Hcutpr1 by omega.
    so (IH1 _#5 Hsum Hlsub Hcutpr1) as (e' & Hcut1' & Hlsub').
    so (IH2 _#5 Hsum Hlsub' Hcutpr2) as (e'' & Hcut2' & Hlsub'').
    exists e''.
    split; eauto using @cutpf; done.
split; [exact Hlp |].
assert (commute_lsub_prop N) as Hl.
  (* commute_lsub *)
  clear IHmr IHml Hr.
  unfold commute_lsub_prop.
  intros z a1 a2 r1 p1 r2 p2 r3 r12 r23 Hsum Hlsubf12 Hlsubf23.
  revert p2 r3 r23 Hsum Hlsubf23.
  induct Hlsubf12.

  (* ret *)
  intros a1 v p1 r r' Hcutp p2 r3 r23 Hsum Hlsubf23.
  #2 exploit (cutpf_sh_cancel a1 v p1 (bound p2) (bound p2) (bound p1) r3) as Hsubr3; try omega.
    eapply cutpf_patmatch; eauto; done.
  replace (bound p1 - bound p1) with 0 in Hsubr3 by omega.
  rewrite -> sh_zero in Hsubr3; [].
  unfold commute_lsub_cutp_prop in Hlp.
  repl (bound p2) with (bound p2 + 0) in Hsubr3 at 1 by omega.
  sor (refine (Hlp _#12 _ Hlsubf23 Hcutp Hsubr3)) as (r'' & Hcut' & Hlsub'); [omega |].
  exists r''.
  split; eauto using @lsubf; done.
  
  (* use *)
  intros a1 h s p1 r2 s' Hlsubf12 IH p2 r3 r23 Hsum Hlsubf23.
  so (IH _#3 Hsum Hlsubf23) as (s'' & Hlsubf1_23 & Hlsubf12_3).
  exists (use h s'').
  split; eauto using @lsubf; done.
  
  (* pm *)
  intros a1 p0 r1 p1 r2 r Hlsubf1 IH p2 r3 r23 Hsum Hlsubf23.
  so (sh_lsubf _ (bound p1) (bound p0) _#4 Hlsubf23) as Hlsubf23'.
  rewrite -> sh_sum_left in Hlsubf23'; [].
  replace (bound p0 + bound p1) with (bound p1 + bound p0) in Hlsubf23' by omega.
  rewrite <- sh_sum in Hlsubf23'; [].
  so (IH _#3 Hsum Hlsubf23') as (r' & Hlsubf1_23 & Hlsubf12_3).
  exists (pm p0 r').
  split; eauto using @lsubf; done.
  
  (* app *)
  intros a1 v s1 p1 r2 s12 Hlsubf12 IH p2 r3 r23 Hsum Hlsubf23.
  so (IH _#3 Hsum Hlsubf23) as (s & Hlsubf1_23 & Hlsubf12_3).
  exists (app v s).
  split; eauto using @lsubf; done.
  
  (* pi1 *)
  intros a1 s1 p1 r2 s12 Hlsubf12 IH p2 r3 r23 Hsum Hlsubf23.
  so (IH _#3 Hsum Hlsubf23) as (s & Hlsubf1_23 & Hlsubf12_3).
  exists (pi1 s).
  split; eauto using @lsubf; done.
  
  (* pi2 *)
  intros a1 s1 p1 r2 s12 Hlsubf12 IH p2 r3 r23 Hsum Hlsubf23.
  so (IH _#3 Hsum Hlsubf23) as (s & Hlsubf1_23 & Hlsubf12_3).
  exists (pi2 s).
  split; eauto using @lsubf; done.
split; [exact Hl |].
split.
  (* commute_cutm_rsub *)
  clear IHmr IHml Hl.
  unfold commute_cutm_rsub_prop.
  intros a b m s r f i m' s' Hsum Hcut Hsubm Hsubs.
  revert b f i m' s' Hsum Hsubm Hsubs.
  induct Hcut.
  
  (* atom *)
  intros a r b f i m' s' Hsum Hsubm Hsubs.
  invert Hsubm; clear Hsubm; [].
  intros r' Hsubr <-.
  invert Hsubs; clear Hsubs; [].
  intros <-.
  exists r'.
  split; eauto using cutmf; done.
  
  (* up *)
  intros a r1 p r2 r Hlsubf b f i m' s' Hsum Hsubm Hsubs.
  invert Hsubm; clear Hsubm; [].
  intros r1' Hsubr1 <-.
  invert Hsubs; clear Hsubs; [].
  intros r2' Hsubr2 <-.
  unfold commute_lsub_rsub_prop in Hlr.
  so (Hlr _#11 Hsum Hlsubf Hsubr1 Hsubr2) as (r' & Hsubr & Hlsubf').
  exists r'.
  split; eauto using cutmf; done.
  
  (* pi *)
  intros a p m v s r m2 Hcutp Hcutm IH c f1 i1 mm ss Hsum Hsublam Hsubapp.
  invert Hsublam; clear Hsublam; [].
  intros m1 Hsubm1 <-.
  invert Hsubapp; clear Hsubapp; [].
  intros v' s' Hsubv Hsubs <-.
  unfold commute_cutp_rsub_prop in Hpr.
  repl i1 with (bound p + i1 - 0 - bound p) in Hsubv by omega.
  sor (refine (Hpr _#12 _ _ Hsubm1 Hcutp Hsubv)) as (m' & Hcutp' & Hsubm2); [omega | omega |].
  clear Hpr.
  repl (bound p + i1 - bound p) with i1 in Hsubm2 by omega.
  so (IH _#5 Hsum Hsubm2 Hsubs) as (r' & Hsubr & Hcutmf').
  exists r'; eauto using cutmf; done.
  
  (* wth1 *)
  intros a m n s r Hcut IH b f i mm ss Hsum Hsubpair Hsubpi.
  invert Hsubpair; clear Hsubpair; [].
  intros m' n' Hsubm Hsubn <-.
  invert Hsubpi; clear Hsubpi; [].
  intros s' Hsubs <-.
  so (IH _#5 Hsum Hsubm Hsubs) as (r' & Hsubr & Hcut').
  exists r'.
  split; eauto using cutmf; done.
  
  (* wth2 *)
  intros a m n s r Hcut IH b f i mm ss Hsum Hsubpair Hsubpi.
  invert Hsubpair; clear Hsubpair; [].
  intros m' n' Hsubm Hsubn <-.
  invert Hsubpi; clear Hsubpi; [].
  intros s' Hsubs <-.
  so (IH _#5 Hsum Hsubn Hsubs) as (r' & Hsubr & Hcut').
  exists r'.
  split; eauto using cutmf; done.

  (* commute_cutm_lsub *)
  clear IHmr IHml Hr Hlr.
  unfold commute_cutm_lsub_prop.
  intros a b m s r1 p r2 s' Hsum Hcut Hlsubf.
  revert p r2 s' Hsum Hlsubf.
  induct Hcut.
  
  (* atom *)
  intros a r p r2 s' Hsum Hlsubf.
  invert Hlsubf; done.
  
  (* up *)
  intros a r1 p1 r2 r12 Hlsubf12 p2 r3 s' Hsum Hlsubf.
  invert Hlsubf; clear Hlsubf; [].
  intros r23 Hlsubf23 <-.
  unfold commute_lsub_prop in Hl.
  so (Hl _#10 Hsum Hlsubf12 Hlsubf23) as (r & Hlsubf1_23 & Hlsubf12_3).
  clear Hl.
  exists r.
  split; eauto using cutmf; done.
  
  (* pi *)
  intros a p1 m v s r m' Hcutp Hcutm IH p2 r2 ss Hsum Hlsubapp.
  invertc Hlsubapp; [].
  intros s' Hlsub <-.
  so (IH _#3 Hsum Hlsub) as (r' & Hlsub' & Hcutm').
  exists r'.
  split; eauto using cutmf; done.
  
  (* wth1 *)
  intros a m n s r Hcutmf IH p r2 ss Hsum Hlsubfpi.
  invert Hlsubfpi; clear Hlsubfpi; [].
  intros s' Hlsubf <-.
  so (IH _#3 Hsum Hlsubf) as (r' & Hlsubf' & Hcutmf').
  exists r'.
  split; eauto using cutmf; done.
  
  (* wth2 *)
  intros a m n s r Hcutmf IH p r2 ss Hsum Hlsubfpi.
  invert Hlsubfpi; clear Hlsubfpi; [].
  intros s' Hlsubf <-.
  so (IH _#3 Hsum Hlsubf) as (r' & Hlsubf' & Hcutmf').
  exists r'.
  split; eauto using cutmf; done.
Qed.


Lemma commute_cutm_rsub :
  forall m s r f i m' s',
    cutm m s r
    -> rsub f i m m'
    -> rsub f i s s'
    -> exists r',
         rsub f i r r'
         /\ cutm m' s' r'.
Proof.
intros m s r f i m' s' Hcut Hsubm Hsubs.
so (fuel_cutm _#3 Hcut) as (a, Hcutf). 
so (fuel_rsub _#4 Hsubm) as (b, Hsubfm).
so (fuel_rsub _#4 Hsubs) as (c, Hsubfs).
so (increase_rsubf _ (b+c) _#4 (le_plus_l _ _) Hsubfm) as Hsubfm'.
so (increase_rsubf _ (b+c) _#4 (le_plus_r _ _) Hsubfs) as Hsubfs'.
so (commutativity (a+(b+c)) anderrrrrl _#9 (eq_refl _) Hcutf Hsubfm' Hsubfs') as (r' & Hsubfm'' & Hsubfs'').
exists r'.
split; eauto using unfuel_rsub, unfuel_cutm.
Qed.


Lemma commute_rsub :
  forall {z} f1 i1 f2 i2 (E E1 E2 : exp z) f2',
    i2 < i1
    -> rsub f1 i1 E E1
    -> rsub f2 i2 E E2
    -> rsub f1 (i1 - i2 - 1) f2 f2'
    -> exists E',
         rsub f2' i2 E1 E'
         /\ rsub f1 (i1-1) E2 E'.
Proof.
intros z f1 i1 f2 i2 E E1 E2 f2' Hleq Hsub1 Hsub2 Hsubf.
so (fuel_rsub _#4 Hsub1) as (a, Hsubf1).
so (fuel_rsub _#4 Hsub2) as (b, Hsubf2).
so (fuel_rsub _#4 Hsubf) as (c, Hsubff).
so (increase_rsubf _ (a+c) _#4 (le_plus_l _ _) Hsubf1) as Hsubf1'.
so (increase_rsubf _ (a+c) _#4 (le_plus_r _ _) Hsubff) as Hsubff'.
so (commutativity ((a+c)+b) andel _#11 (eq_refl _) Hleq Hsubf1' Hsubf2 Hsubff') as (E' & Hsubf1'' & Hsubf2').
exists E'.
split; eauto using unfuel_rsub.
Qed.


Lemma commute_cutp_rsub :
  forall z v p i1 f i2 (E E1 E2 : exp z) v',
    i2 + bound p <= i1
    -> rsub f i1 E E1
    -> cutp v p i2 E E2
    -> rsub f (i1 - i2 - bound p) v v'
    -> exists E',
         cutp v' p i2 E1 E'
         /\ rsub f (i1 - bound p) E2 E'.
Proof.
intros z v p i1 f i2 e e1 e2 v' Hleq Hrsub Hcutp Hsubv.
so (fuel_rsub _#4 Hrsub) as (a, Hrsubf).
so (fuel_cutp _#6 Hcutp) as (b, Hcutpf).
so (fuel_rsub _#4 Hsubv) as (c, Hsubfv).
so (increase_rsubf _ (a+c) _#4 (le_plus_l _ _) Hrsubf) as Hrsubf'.
so (increase_rsubf _ (a+c) _#4 (le_plus_r _ _) Hsubfv) as Hsubfv'.
so (commutativity ((a+c)+b) anderl _#12 (eq_refl _) Hleq Hrsubf' Hcutpf Hsubfv') as (e' & Hcutpf' & Hrsubf'').
exists e'.
split; eauto using unfuel_rsub, unfuel_cutp.
Qed.


Lemma commute_cutp_cutp :
  forall z v1 p1 i1 v2 p2 i2 (E E1 E2 : exp z) v2',
    i2 + bound p2 <= i1
    -> cutp v1 p1 i1 E E1
    -> cutp v2 p2 i2 E E2
    -> cutp v1 p1 (i1 - i2 - bound p2) v2 v2'
    -> exists E',
         cutp v2' p2 i2 E1 E'
         /\ cutp v1 p1 (i1 - bound p2) E2 E'.
Proof.
intros z v1 p1 i1 v2 p2 i2 E E1 E2 v2' Hleq Hcut1 Hcut2 Hcutv2.
revert v2 p2 i2 E2 v2' Hleq Hcut2 Hcutv2.
induct Hcut1.

(* down *)
intros z m i e e1 Hsub1 v2 p2 i2 e2 v2' Hleq Hcut2 Hcutv2.
invertc Hcutv2; [].
intros Hsubv2.
so (commute_cutp_rsub _#10 Hleq Hsub1 Hcut2 Hsubv2) as (e' & Hcut & Hsub).
eauto using @cutp; done.

(* one *)
intros z i e1 v2 p2 i2 e2 v2' Hleq Hcut2 Hcutv2.
invertc Hcutv2; [].
intros <-.
eauto using @cutp; done.

(* tens *)
intros z v w p1 q1 i e e1a e1 _ IH1 _ IH2 v2 p2 i2 e2 v2'' Hleq Hcut2 Hcutv2.
invertc Hcutv2; [].
intros v2' Hcutv2' Hcutv2''.
repl (bound q1 + (i - i2 - bound p2)) with (bound q1 + i - i2 - bound p2) in Hcutv2' by omega.
sor (refine (IH1 _#5 _ Hcut2 Hcutv2')) as (e1a' & Hcut1a & Hcut2a); [omega |].
so (IH2 _#5 Hleq Hcut1a Hcutv2'') as (e' & Hcut1b & Hcut2b).
exists e'.
split; eauto; [].
eapply cutp_tens; eauto; [].
repl (bound q1 + (i - bound p2)) with (bound q1 + i - bound p2) by omega.
assumption.
Qed.


(* Lowering *)

(* The smallest closed term. *)
Definition bogus : term := shiftp (ret unitp).

Lemma rsub_bogus :
  forall f i, rsub f i bogus bogus.
Proof.
intros f i.
eauto using @rsub.
Qed.


Lemma dummy_sub :
  forall j,
    exists v p,
      patmatch v p
      /\ bound p = j
      /\ forall f i, rsub f i v v.
Proof.
intros j.
induct j.

(* nil *)
exists unitp, unitpat.
do2 2 split; eauto using patmatch, @rsub; done.

(* cons *)
intros j (v & p & Hmatch & Hlen & Hsub).
exists (pairp v (thunk bogus)), (pairpat p shiftm).
do2 2 split; eauto using patmatch, @rsub; [].
simpl.
omega.
Qed.


Lemma closed_rsub_to_cutp :
  forall z (e : exp z),
    (forall f i, rsub f i e e)
    -> (forall v p i, patmatch v p -> cutp v p i e e).
Proof.
intros z e Hstat v p i Hmatch.
revert i.
induct Hmatch; eauto using @cutp.
Qed.


Lemma lower_eq :
  forall {z} i j (e e' : exp z),
    sh i j e = sh i j e'
    -> e = e'.
Proof.
intros z i j e e' Heq.
so (dummy_sub j) as (v & p & Hmatch & Hlen & _).
so (cutp_sh_cancel_all v p i e Hmatch) as Hlower.
so (cutp_sh_cancel_all v p i e' Hmatch) as Hlower'.
rewrite <- Hlen in Heq; [].
rewrite -> Heq in Hlower; [].
exact (cutp_fun _#6 Hlower Hlower').
Qed.


Lemma lower_sub :
  forall {z} i i' j f (e e' : exp z),
    rsub f (i+i'+j) (sh i' j e) e'
    -> exists e'', rsub f (i+i') e e''.
Proof.
intros z i i' j f e e' Hsub1.
so (dummy_sub j) as (v & p & Hmatch & Hlen & Hsubg).
so (cutp_sh_cancel_all v p i' e Hmatch) as Hsub2.
rewrite -> Hlen in Hsub2; [].
sor (refine (commute_cutp_rsub _#10 _ Hsub1 Hsub2 (Hsubg _ _))) as H; [omega |].
destruct H as (e'' & _ & Hsub).
rewrite -> Hlen in Hsub; [].
repl (i + i' + j - j) with (i + i') in Hsub by omega.
eexists; eauto.
Qed.


Lemma rsub_of_sh :
  forall {z} i i' j f (e e' : exp z),
    rsub f (i + i' + j) (sh i' j e) e'
    -> exists e'',
         rsub f (i + i') e e''
         /\ e' = sh i' j e''.
Proof.
intros z i i' j f e e' Hsub.
so (lower_sub _#6 Hsub) as (e'' & Hsub').
sor (refine (rsub_sh_commute_right _ _ i' j _ _ _ Hsub')) as Hsub''; [omega |].
so (rsub_fun _#5 Hsub Hsub''); subst e'.
exists e''.
auto.
Qed.


Lemma lower_cutp :
  forall {z} i i' j v p (e e' : exp z),
    cutp v p (i+i'+j) (sh i' j e) e'
    -> exists e'', cutp v p (i+i') e e''.
Proof.
intros z i i' j v p e e' Hsub1.
so (dummy_sub j) as (w & q & Hmatch & Hlen & Hsubw).
so (cutp_sh_cancel_all w q i' e Hmatch) as Hsub2.
rewrite -> Hlen in Hsub2; [].
so (closed_rsub_to_cutp _ _ Hsubw) as Hcutw.
sor (refine (commute_cutp_cutp _#11 _ Hsub1 Hsub2 (Hcutw _ _ _ (cutp_patmatch _#5 Hsub1)))) as H; [omega |].
destruct H as (e'' & _ & Hsub).
rewrite -> Hlen in Hsub; [].
repl (i + i' + j - j) with (i + i') in Hsub by omega.
eexists; eauto.
Qed.


Lemma cutp_of_sh :
  forall {z} i i' j v p (e e' : exp z),
    cutp v p (i + i' + j) (sh i' j e) e'
    -> exists e'',
         cutp v p (i + i') e e''
         /\ e' = sh i' j e''.
Proof.
intros z i i' j v p e e' Hsub.
so (lower_cutp _#7 Hsub) as (e'' & Hsub').
sor (refine (cutp_sh_commute_right _ _ _ i' j _ _ _ Hsub')) as Hsub''; [omega |].
so (cutp_fun _#6 Hsub Hsub''); subst e'.
exists e''.
auto.
Qed.


Lemma lower_cutp_left :
  forall {z} i i' j v p (e e': exp z),
    cutp (sh i' j v) p i (sh (i + i' + bound p) j e) e'
    -> exists e'',
         cutp v p i e e''.
Proof.
intros z i i' j v p e e' Hsub1.
so (dummy_sub j) as (w & q & Hmatch & Hlen & _).
so (cutp_sh_cancel_all w q i' v Hmatch) as Hsubv.
so (cutp_sh_cancel_all w q (i + i' + bound p) e Hmatch) as Hsub2.
rewrite -> Hlen in Hsubv; [].
rewrite -> Hlen in Hsub2; [].
#3 sor (refine (commute_cutp_cutp _#11 _ Hsub2 Hsub1 _)) as H.
  omega.

  repl (i + i' + bound p - i - bound p) with i' by omega.
  exact Hsubv.
destruct H as (e'' & Hsub & _).
eauto.
Qed.


Lemma cutp_of_sh_left :
  forall {z} i i' j v p (e e': exp z),
    cutp (sh i' j v) p i (sh (i + i' + bound p) j e) e'
    -> exists e'',
         cutp v p i e e''
         /\ e' = sh (i+i') j e''.
Proof.
intros z i i' j v p e e' Hsub.
so (lower_cutp_left _#7 Hsub) as (e'' & Hsub').
so (cutp_sh_commute_left _#4 i' j _ _ Hsub') as Hsub''.
repl (i + bound p + i') with (i + i' + bound p) in Hsub'' by omega.
so (cutp_fun _#6 Hsub Hsub''); subst e'.
exists e''.
auto.
Qed.


Lemma rsub_sh_uncommute_right :
  forall {z} f i i' j (e e' : exp z),
    i' <= i
    -> rsub f (i+j) (sh i' j e) (sh i' j e')
    -> rsub f i e e'.
Proof.
intros z f i i' j e e' Hleq Hsub.
repl (i+j) with ((i-i')+i'+j) in Hsub by omega.
so (lower_sub _#6 Hsub) as (e'' & Hsub').
repl ((i-i')+i'+j) with (i+j) in Hsub by omega.
repl (i-i'+i') with i in Hsub' by omega.
so (rsub_sh_commute_right _ _ i' j _ _ Hleq Hsub') as Hsub''.
so (rsub_fun _#5 Hsub Hsub'') as Heq.
so (lower_eq _#4 Heq); subst e''.
exact Hsub'.
Qed.


(* Limited inversion and compatibility for cutp *)

Lemma cutp_invert1_sh :
  forall {z1 z2} (c : exp z1 -> exp z2) j,
    (forall f i d e, rsub f i (c d) e -> exists d', e = c d' /\ rsub f (j + i) d d')
    -> forall v p i d e,
         cutp v p i (c d) e
         -> exists d',
              e = c d'
              /\ cutp v p (j + i) d d'.
Proof.
intros z1 z2 c j Hinvert v p i d e Hcut.
revert v i d e Hcut.
induct p.

(* down *)
intros v i d e Hcut.
invertc Hcut; [].
intros m Hsub <-.
so (Hinvert _#4 Hsub) as (d' & -> & Hsub').
exists d'.
eauto using @cutp; done.

(* one *)
intros v i d e Hcut.
invertc Hcut; [].
intros <- <-.
exists d.
eauto using @cutp; done.

(* tens *)
intros p IH1 q IH2 vv i d e Hcut.
invertc Hcut; [].
intros v w e' Hcut1 Hcut2 <-.
so (IH1 _#4 Hcut1) as (d' & -> & Hcut1').
so (IH2 _#4 Hcut2) as (d'' & -> & Hcut2').
exists d''.
split; auto; [].
#2 eapply cutp_tens.
  repl (bound q + (j + i)) with (j + (bound q + i)) by omega.
  eauto; done.

  eauto; done.
Qed.


Lemma cutp_invert1 :
  forall {z1 z2} (c : exp z1 -> exp z2),
    (forall f i d e, rsub f i (c d) e -> exists d', e = c d' /\ rsub f i d d')
    -> forall v p i d e,
         cutp v p i (c d) e
         -> exists d',
              e = c d'
              /\ cutp v p i d d'.
Proof.
intros z1 z2 c.
so (cutp_invert1_sh c 0).
eauto.
Qed.


Lemma cutp_invert2 :
  forall {z1 z2 z3} (c : exp z1 -> exp z2 -> exp z3),
    (forall f i d1 d2 e,
       rsub f i (c d1 d2) e
       -> exists d1' d2', e = c d1' d2' /\ rsub f i d1 d1' /\ rsub f i d2 d2')
    -> forall v p i d1 d2 e,
         cutp v p i (c d1 d2) e
         -> exists d1' d2',
              e = c d1' d2'
              /\ cutp v p i d1 d1'
              /\ cutp v p i d2 d2'.
Proof.
intros z1 z2 z3 c Hinvert v p i d1 d2 e Hcut.
revert v i d1 d2 e Hcut.
induct p.

(* down *)
intros v i d1 d2 e Hcut.
invertc Hcut; [].
intros m Hsub <-.
so (Hinvert _#5 Hsub) as (d1' & d2' & -> & Hsub1 & Hsub2).
exists d1', d2'.
eauto using @cutp; done.

(* one *)
intros v i d1 d2 e Hcut.
invertc Hcut; [].
intros <- <-.
exists d1, d2.
eauto using @cutp; done.

(* tens *)
intros p IH1 q IH2 vv i d1 d2 e Hcut.
invertc Hcut; [].
intros v w e' Hcut1 Hcut2 <-.
so (IH1 _#5 Hcut1) as (d1' & d2' & -> & Hcut1' & Hcut2').
so (IH2 _#5 Hcut2) as (d1'' & d2'' & -> & Hcut1'' & Hcut2'').
exists d1'', d2''.
eauto using @cutp; done.
Qed.


Lemma cutp_compat2 :
  forall {z1 z2 z} v p i (e : exp z1 -> exp z2 -> exp z) e1 e2 e1' e2',
    cutp v p i e1 e1'
    -> cutp v p i e2 e2'
    -> (forall f j d1 d2 d1' d2', rsub f j d1 d1' -> rsub f j d2 d2' -> rsub f j (e d1 d2) (e d1' d2'))
    -> cutp v p i (e e1 e2) (e e1' e2').
Proof.
intros z1 z2 z v p i e e1 e2 e1' e2' Hcut1 Hcut2 Hcompat.
revert z1 z2 z v i e e1 e2 e1' e2' Hcut1 Hcut2 Hcompat.
induct p.

(* down *)
intros z1 z2 z v i e e1 e2 e1' e2' Hcut1 Hcut2 Hcompat.
invertc Hcut1; [].
intros m Hsub1 <-.
invertc Hcut2; [].
intros Hsub2.
apply cutp_down; eauto; done.

(* one *)
intros z1 z2 z v i e e1 e2 e1' e2' Hcut1 Hcut2 Hcompat.
invertc Hcut1; [].
intros <- <-.
invertc Hcut2; [].
intros <-.
apply cutp_one; done.

(* tens *)
intros p IH1 q IH2 z1 z2 z vv i e e1 e2 e1'' e2'' Hcut1 Hcut2 Hcompat.
invertc Hcut1; [].
intros v w e1' Hcut1' Hcut1'' <-.
invertc Hcut2; [].
intros e2' Hcut2' Hcut2''.
eapply cutp_tens; eauto; done.
Qed.


Lemma cutp_compat1 :
  forall {z1 z} v p i (e : exp z1 -> exp z) e1 e1',
    cutp v p i e1 e1'
    -> (forall f j d1 d1', rsub f j d1 d1' -> rsub f j (e d1) (e d1'))
    -> cutp v p i (e e1) (e e1').
Proof.
intros z1 z v p i e e1 e1' Hcut Hcompat.
exploit (cutp_compat2 v p i (fun d (_ : exp z1) => e d) e1 e1 e1' e1'); auto.
Qed.


Lemma cutp_compat0 :
  forall {z} v p i (e : exp z),
    (forall f j, rsub f j e e)
    -> patmatch v p
    -> cutp v p i e e.
Proof.
intros z v p i e Hsub Hmatch.
revert i e Hsub.
induct Hmatch; intros; eauto using @cutp.
Qed.


(* Contexts *)

Lemma shctx_zero :
  forall i G,
    shctx i 0 G = G.
Proof.
intros i G.
revert i.
induct G; intros; simpl; f_equal; eauto using sh_zero.
Qed.


Lemma shctx_commute :
  forall i j i' j' G,
    shctx (i + j + i') j' (shctx i j G) = shctx i j (shctx (i+i') j' G).
Proof.
intros i j i' j' G.
induct G.

(* nil *)
auto; done.

(* cons *)
intros A G IH.
simpl.
f_equal; auto; [].
rewrite -> !shctx_length; [].
repl (i + i' + length G) with ((i + length G) + i') by omega.
rewrite <- sh_commute; [].
f_equal; omega.
Qed.


Lemma rsubctx_sh_commute_right_0 :
  forall f i j G G',
    rsubctx f i G G'
    -> rsubctx f (i+j) (shctx 0 j G) (shctx 0 j G').
Proof.
intros f i j G G' H.
induct H.

(* nil *)
intros f i.
apply rsubctx_nil; done.

(* cons *)
intros f i G A G' A' HsubG IH Hsub.
simpl.
apply rsubctx_cons; auto; [].
rewrite -> shctx_length; [].
rewrite <- (rsubctx_length _#4 HsubG); [].
replace (i + j + length G) with ((i + length G) + j) by omega.
apply rsub_sh_commute_right; auto; omega.
Qed.


Lemma shctx_append :
  forall i j k G1 G2,
    length G1 = k
    -> shctx i j (G1;; G2) = shctx i j G1;; shctx (i+k) j G2.
Proof.
intros i j k G1 G2 Hlen.
induct G2.

(* nil *)
auto; done.

(* cons *)
intros A G IH.
simpl.
f_equal; auto; [].
rewrite -> app_length; [].
f_equal; omega.
Qed.


Lemma rsubctx_append :
  forall f i j G1 G2 G1' G2',
    length G1 = j
    -> rsubctx f i G1 G1'
    -> rsubctx f (i+j) G2 G2'
    -> rsubctx f i (G1;; G2) (G1';; G2').
Proof.
intros f i j G1 G2 G1' G2' Hlen Hsub1 Hsub2.
remember (i+j) as k eqn:Heqk.
revert Heqk Hsub1.
induct Hsub2.

(* nil *)
intros f ? -> Hsub.
exact Hsub.

(* cons *)
intros f ? G2 A G' A' H IH Hrsub -> Hsub1.
simpl.
apply rsubctx_cons; auto; [].
rewrite -> app_length; [].
force_exact Hrsub.
f_equal; omega.
Qed.


Lemma cutpctx_nil :
  forall v p i,
    patmatch v p
    -> cutpctx v p i nil nil.
Proof.
intros v p i Hmatch.
revert i.
induct Hmatch; intros; eauto using cutpctx, rsubctx.
Qed.


Lemma cutpctx_append :
  forall v p i G1 G1' G2 G2',
    cutpctx v p i G1 G1'
    -> cutpctx v p (i + length G1) G2 G2'
    -> cutpctx v p i (G1;; G2) (G1';; G2').
Proof.
intros v p i G1 G1' G2 G2' Hcut1 Hcut2.
revert G2 G2' Hcut2.
induct Hcut1;
try (intros; invertc Hcut2; intros; subst; eauto using cutpctx, rsubctx_append; done).
(* tens *)
intros v w p q i G1 G1' G1'' HcutG1 IH1 HcutG1' IH2 G2 G2'' Hcut2.
invertc Hcut2; [].
intros G2' HcutG2 HcutG2'.
repl (bound q + (i + length G1)) with (bound q + i + length G1) in HcutG2 by omega.
so (IH1 _ _ HcutG2) as Hcut.
rewrite (cutpctx_length _#5 HcutG1) in HcutG2'.
so (IH2 _ _ HcutG2') as Hcut'.
eapply cutpctx_tens; eauto; done.
Qed.


Lemma rsubctx_shctx_cancel :
  forall f i i' j G,
    i >= i'
    -> i < i' + j
    -> rsubctx f i (shctx i' j G) (shctx i' (j - 1) G).
Proof.
intros f i i' j G Hlower Hupper.
revert i i' j Hlower Hupper.
induct G.

(* nil *)
intros; simpl; apply rsubctx_nil; done.

(* cons *)
intros A G IH i i' j Hlower Hupper.
simpl.
apply rsubctx_cons; eauto; [].
rewrite -> shctx_length; [].
apply rsub_sh_cancel; omega.
Qed.


Lemma cutpctx_shctx_cancel :
  forall v p i i' j G,
    i >= i'
    -> i + bound p <= i' + j
    -> patmatch v p
    -> cutpctx v p i (shctx i' j G) (shctx i' (j - bound p) G).
Proof.
intros v p i i' j G Hlower Hupper Hmatch.
revert i i' j G Hlower Hupper.
induct Hmatch.

(* down *)
simpl; intros; apply cutpctx_down; apply rsubctx_shctx_cancel; omega.

(* one *)
intros i i' j G Hlower Hupper.
simpl.
repl (j - 0) with j by omega.
apply cutpctx_one; done.

(* tens *)
intros v w p q _ IH1 _ IH2 i i' j G Hlower Hupper.
simpl in Hupper |- *.
exploit (IH1 (bound q + i) i' j G) as Hcut1; try omega; [].
exploit (IH2 i i' (j - bound p) G) as Hcut2; try omega; [].
eapply cutpctx_tens; eauto; [].
repl (j - (bound p + bound q)) with (j - bound p - bound q) by omega.
exact Hcut2.
Qed.


Lemma cutpctx_shctx_cancel_all :
  forall v p i G,
    patmatch v p
    -> cutpctx v p i (shctx i (bound p) G) G.
Proof.
intros v p i G Hmatch.
rewrite <- (shctx_zero i G) at 2.
repl 0 with (bound p - bound p) by omega.
apply cutpctx_shctx_cancel; auto; omega.
Qed.


Lemma cutpctx_patmatch :
   forall v p i G G',
     cutpctx v p i G G'
     -> patmatch v p.
Proof.
intros v p i G G' H.
induct H; intros; eauto using patmatch.
Qed.
