
Require Import Arith.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Substitution.
Require Import Self.
Require Import Static.


(* Inst *)

Inductive inst : term -> nat -> res -> res -> Prop :=
| inst_use {f i h s h' s'}
    : rsub_head i h h'
      -> rsub f i s s'
      -> inst f i (use h s) (use h' s')

| inst_ret {f i v v'}
    : rsub f i v v'
      -> inst f i (ret v) (ret v')
.


Lemma inst_impl_rsub :
  forall f i r r',
    inst f i r r' -> rsub f i r r'.
Proof.
intros f i r r' H.
induct H; intros; eauto using @rsub.
Qed.


Lemma inst_sh_cancel :
  forall f i i' j r,
    i >= i'
    -> i < i' + j
    -> inst f i (sh i' j r) (sh i' (j-1) r).
Proof.
intros f i i' j r Hlower Hupper.
#2 casesdep r.

(* ret *)
intros v.
simpl.
apply inst_ret; [].
apply rsub_sh_cancel; auto; done.

(* use *)
intros h s.
so (rsub_sh_cancel f _#3 (use h s) Hlower Hupper) as Hsub.
simpl in Hsub.
invertc Hsub; eauto using inst; [].
intros ? _ _ Heq.
exfalso.
destruct h; simpl in Heq; try (discriminate Heq); [].
injection Heq; clear Heq; [].
intros Heq'.
unfold sh_var in Heq'.
forget (lt_dec n i') as b.
destruct b; omega.
Qed.


Lemma inst_sh_cancel_all :
  forall f i (r : res),
    inst f i (sh i 1 r) r.
Proof.
intros f i r.
exploit (inst_sh_cancel f i i 1 r) as H; try omega; [].
repl (1 - 1) with 0 in H by omega.
rewrite -> sh_zero in H; [].
exact H.
Qed.


(* Mostly copied from the inst case of rsubf_sh_commute_right *)
Lemma inst_sh_commute_right :
  forall f i i' j r r',
    i' <= i
    -> inst f i r r'
    -> inst f (i+j) (sh i' j r) (sh i' j r').
Proof.
intros f i i' j r r' Hleq Hinst.
revert Hleq.
#2 cases Hinst.

(* use *)
intros f i h s h' s' Hsubh Hsubs Hleq.
simpl.
apply inst_use; eauto using @rsub_sh_commute_right; [].
#3 invert Hsubh; clear Hsubh.
  intros c <- <-.
  apply rsub_const; done.

  intros k Hlt <- <-.
  apply rsub_var_lt; [].
  unfold sh_var.
  forget (lt_dec k i') as b; destruct b; omega.

  intros k Hgeq <- <-.
  simpl.
  unfold sh_var.
  forget (lt_dec k i') as b; forget (lt_dec (k - 1) i') as b'.
  destruct b as [H | H]; destruct b' as [H' | H']; try omega; [].
  replace (j + (k - 1)) with ((j + k) - 1) by omega.
  apply rsub_var_gt; omega.

(* ret *)
intros f i v v' Hsubv Hleq.
simpl.
apply inst_ret; [].
apply rsub_sh_commute_right; auto.
Qed.


(* Composition *)

Lemma compose_fun :
  forall {z} (e : exp z) s e1 e2,
    compose e s e1
    -> compose e s e2
    -> e1 = e2.
Proof.
intros z e s e1 e2 H Hcompose.
revert e2 Hcompose.
induct H; intros; invertc Hcompose; intros; subst; f_equal; eauto.
Qed.


Lemma sh_compose :
  forall {z} i j (s1 : exp z) s2 s,
    compose s1 s2 s
    -> compose (sh i j s1) (sh i j s2) (sh i j s).
Proof.
intros z i j s1 s2 s H.
revert i j.
induct H; try (intros; simpl; eauto using @compose; done); [].

(* pm *)
intros p r1 s2 r _ IH i j.
simpl.
so (IH (bound p + i) j) as H.
apply compose_pm; [].
repl (bound p + i) with (0 + bound p + i) in H at 2 by omega.
rewrite -> sh_commute in H; [].
repl (0 + i) with i in H by omega.
exact H.
Qed.


Lemma compose_of_sh :
  forall {z} i j (s1 : exp z) s2 s,
    compose (sh i j s1) (sh i j s2) s
    -> exists s',
         compose s1 s2 s'
         /\ s = sh i j s'.
Proof.
intros z i j s1 s2 s Hcompose.
remember (sh i j s1) as t1 eqn:Heqt1.
remember (sh i j s2) as t2 eqn:Heqt2.
revert i j s1 s2 Heqt1 Heqt2.
induct Hcompose.

(* nl *)
intros s i j s1 s2 Heq ->.
revert Heq.
casesdep s1; try (intros; discriminate Heq); [].
intros _.
exists s2.
auto using @compose; done.

(* pm *)
intros p r1 ? r Hcompose IH i j rr s Heq ->.
revert Heq.
casesdep rr; try (intros; discriminate Heq); [].
intros q r1' Heq.
injection Heq; clear Heq.
intros -> <-.
#2 exploit (IH (bound p + i) j r1' (sh 0 (bound p) s)) as H; auto.
  repl (bound p + i) with (0 + (bound p) + i) by omega.
  rewrite -> sh_commute; [].
  reflexivity.
destruct H as (s' & Hcompose' & ->).
exists (pm p s').
simpl.
split; auto using @compose; done.

(* app *)
intros v s1 ? s Hcompose IH i j ss s2 Heq ->.
revert Heq.
casesdep ss; try (intros; discriminate Heq); [].
intros v' s1' Heq.
injection Heq; clear Heq; [].
intros -> ->.
so (IH _#4 (eq_refl _) (eq_refl _)) as (s' & Hcompose' & ->).
exists (app v' s').
split; auto using @compose; done.

(* pi1 *)
intros s1 ? s Hcompose IH i j ss s2 Heq ->.
revert Heq.
casesdep ss; try (intros; discriminate Heq); [].
intros s1' Heq.
injection Heq; clear Heq; [].
intros ->.
so (IH _#4 (eq_refl _) (eq_refl _)) as (s' & Hcompose' & ->).
exists (pi1 s').
split; auto using @compose; done.

(* pi2 *)
intros s1 ? s Hcompose IH i j ss s2 Heq ->.
revert Heq.
casesdep ss; try (intros; discriminate Heq); [].
intros s1' Heq.
injection Heq; clear Heq; [].
intros ->.
so (IH _#4 (eq_refl _) (eq_refl _)) as (s' & Hcompose' & ->).
exists (pi2 s').
split; auto using @compose; done.

(* use *)
intros h s1 ? s Hcompose IH i j ss s2 Heq ->.
revert Heq.
casesdep ss; try (intros; discriminate Heq); [].
intros h' s1' Heq.
injection Heq; clear Heq; [].
intros -> ->.
so (IH _#4 (eq_refl _) (eq_refl _)) as (s' & Hcompose' & ->).
exists (use h' s').
split; auto using @compose; done.
Qed.


Lemma compose_assoc :
  forall {z} (s1 : exp z) s2 s3 s12 s,
    compose s1 s2 s12
    -> compose s12 s3 s
    -> exists s23,
         compose s2 s3 s23
         /\ compose s1 s23 s.
Proof.
intros z s1 s2 s3 s12 s Hcompose12 Hcompose12_3.
revert s3 s Hcompose12_3.
induct Hcompose12.

(* nl *)
intros s2 s3 s Hcompose12_3.
exists s; auto using @compose; done.

(* pm *)
intros p r1 s2 r12 _ IH s3 s Hcomposepm.
invertc Hcomposepm; [].
intros r123 Hcompose12_3 <-.
so (IH _ _ Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
so (compose_of_sh _#5 Hcompose23) as (s23' & Hcompose23' & ->).
exists s23'.
split; auto using @compose; done.

(* app *)
intros v s1 s2 s12 ? IH s3 s Hcomposeapp.
invertc Hcomposeapp; [].
intros s123 Hcompose12_3 <-.
so (IH _ _ Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
exists s23.
split; auto using @compose; done.

(* pi1 *)
intros s1 s2 s12 _ IH s3 s Hcompose.
invertc Hcompose; [].
intros s123 Hcompose12_3 <-.
so (IH _ _ Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
exists s23.
split; auto using @compose; done.

(* pi2 *)
intros s1 s2 s12 _ IH s3 s Hcompose.
invertc Hcompose; [].
intros s123 Hcompose12_3 <-.
so (IH _ _ Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
exists s23.
split; auto using @compose; done.

(* use *)
intros h s1 s2 s12 _ IH s3 s Hcompose.
invertc Hcompose; [].
intros s123 Hcompose12_3 <-.
so (IH _ _ Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
exists s23.
split; auto using @compose; done.
Qed.


(* Negative spines/results *)

Inductive nspine : spine -> Prop :=
| nspine_nl
    : nspine nl

| nspine_app {v s}
    : nspine s
      -> nspine (app v s)

| nspine_pi1 {s}
    : nspine s
      -> nspine (pi1 s)

| nspine_pi2 {s}
    : nspine s
      -> nspine (pi2 s).


Inductive nres : res -> Prop :=
| nres_use {h s}
    : nspine s
      -> nres (use h s)
.


Lemma sh_nspine :
  forall i j s,
    nspine s
    -> nspine (sh i j s).
Proof.
intros i j s H.
induct H; intros; simpl; eauto using nspine.
Qed.


Lemma sh_nres :
  forall i j r,
    nres r
    -> nres (sh i j r).
Proof.
intros i j r H.
cases H.
intros h s His.
simpl.
apply nres_use; eauto using sh_nspine.
Qed.


Lemma rsub_nspine :
  forall f i s s',
    rsub f i s s'
    -> nspine s
    -> nspine s'.
Proof.
intros f i s s' Hsub His.
revert f i s' Hsub.
induct His; intros; invertc Hsub; intros; subst; eauto using nspine.
Qed.


Lemma cutp_nspine :
  forall v p i s s',
    cutp v p i s s'
    -> nspine s
    -> nspine s'.
Proof.
intros v p i s s' Hcut His.
revert v i s s' Hcut His.
induct p; intros; invertc Hcut; intros; subst; eauto using nspine, rsub_nspine.
Qed.


Lemma compose_nspine :
  forall s1 s2 s,
    compose s1 s2 s
    -> nspine s1
    -> nspine s2
    -> nspine s.
Proof.
intros s1 s2 s Hcompose His1 His2.
revert s2 s Hcompose His2.
induct His1; intros; invertc Hcompose; intros; subst; eauto using nspine.
Qed.


Lemma compose_nres_nspine :
  forall r s r',
    compose r s r'
    -> nres r
    -> nspine s
    -> nres r'.
Proof.
intros r s2 r' Hcompose His Hiss.
revert s2 r' Hcompose Hiss.
induct His; intros; invert Hcompose; intros; subst; eauto using nres, compose_nspine.
Qed.


Lemma sof_nspine :
  forall G s A B,
    sof G s A (susp B)
    -> nspine s.
Proof.
intros G s A B H.
remember (susp B) as U eqn:HeqU.
revert B HeqU.
induct H; intros; subst; try discriminate; eauto using nspine.
Qed.



Lemma rof_nres :
  forall G r A,
    rof G r (susp A)
    -> nres r.
Proof.
intros G r A Hof.
invert Hof; [].
intros h s B _ Hofs <-.
apply nres_use; [].
eapply sof_nspine; eauto.
Qed.


Lemma nspine_compose_nl :
  forall s,
     nspine s
     -> compose s nl s.
Proof.
intros s H.
induct H; intros; eauto using @compose.
Qed.


Lemma nres_compose_nl :
  forall r,
    nres r
    -> compose r nl r.
Proof.
intros s H.
induct H; intros; eauto using @compose, nspine_compose_nl.
Qed.


Lemma can_compose_nspine :
  forall s1 s2,
    nspine s1
    -> exists s, compose s1 s2 s.
Proof.
intros s1 s2 His.
induct His;
  intros;
  repeat
    (match goal with
     | H : exists _,_ |- _ => destruct H
     end);
  eexists;
  eauto using @compose.
Qed.


Lemma can_compose_nres :
  forall r s,
    nres r
    -> exists r', compose r s r'.
Proof.
intros r s2 His.
induct His.
intros h s1 His'.
so (can_compose_nspine _ s2 His') as (s & Hcompose).
exists (use h s).
apply compose_use; auto.
Qed.


(* Rigidity *)

Inductive rigid : nat -> nat -> res -> Prop :=
| rigid_const {i j c s}
    : rigid i j (use (const c) s)

| rigid_var {i j k s}
    : (k < i \/ k >= i + j)
      -> rigid i j (use (var k) s)
.


Lemma sh_rigid :
  forall i j k r,
    rigid i j r
    -> rigid (k+i) j (sh 0 k r).
Proof.
intros i j k r Hrigid.
cases Hrigid; eauto using rigid; [].
intros i j l s Hineq.
simpl.
unfold sh_var.
forget (lt_dec l 0) as b; destruct b; try omega; [].
apply rigid_var; omega.
Qed.


Lemma sh_implies_rigid :
  forall i j k r,
    i + j <= k
    -> nres r
    -> rigid i j (sh 0 k r).
Proof.
intros i j k r Hleq His.
destruct His as [h s Hiss].
destruct h as [c | l].
  simpl.
  apply rigid_const; done.

  simpl.
  apply rigid_var; [].
  unfold sh_var.
  forget (lt_dec l 0) as b; destruct b; omega.
Qed.


Lemma rigid_adjust :
  forall i j i' j' r,
    i <= i'
    -> i' + j' <= i + j
    -> rigid i j r
    -> rigid i' j' r.
Proof.
intros i j i' j' r Hlower Hupper Hrigid.
#2 invert Hrigid.
  intros; subst; eauto using rigid; done.

  intros k s Hineq <-.
  apply rigid_var; omega.
Qed.


Lemma rigid_compose :
  forall i j r s r',
    compose r s r'
    -> rigid i j r
    -> rigid i j r'.
Proof.
intros i j r s r' Hcompose Hrigid.
invertc Hcompose; [].
intros h s1 s2 Hcompose <- <-.
invert Hrigid; intros; subst; eauto using rigid.
Qed.


Lemma rsub_rigid_invert :
  forall f i r r' j k,
    rsub f i r r'
    -> rigid j k r
    -> j <= i
    -> i < j+k
    -> inst f i r r'.
Proof.
intros f i r r' j k Hsub Hrigid Hlower Hupper.
#2 invertc Hrigid.
  intros c s <-.
  invert Hsub; intros; subst; eapply inst_use; eauto; done.

  intros l s Hineq <-.
  assert (l <> i) as Hneq by omega.
  invert Hsub; intros; subst; try omega; eauto using inst; done.
Qed.


(* We don't currently use this in any significant way, so it's a good candidate for cleanup. *)
Inductive rsubs_head : nat -> nat -> head -> head -> Prop :=
| rsubs_const {i j c}
    : rsubs_head i j (const c) (const c)

| rsubs_var_lt {i j k}
    : k < i
      -> rsubs_head i j (var k) (var k)

| rsubs_var_gt {i j k}
    : k >= i + j
      -> rsubs_head i j (var k) (var (k-j))
.


Lemma rsubs_head_rigid :
  forall i j i' j' h h' s s',
    i <= i'
    -> i' + j' <= i + j
    -> rigid i j (use h s)
    -> rsubs_head i' j' h h'
    -> rigid i (j - j') (use h' s').
Proof.
intros i j i' j' h h' s s' Hlower Hupper Hrigid Hsubh.
#2 invertc Hrigid.
  intros c <-.
  invertc Hsubh; [].
  intros <-; apply rigid_const; done.
intros k Hineq <-.
invert Hsubh; intros; subst; apply rigid_var; omega.
Qed.


Lemma cutp_rigid_omnibus :
  forall v p i h s r2 j k,
    j <= i
    -> i + bound p <= j + k
    -> rigid j k (use h s)
    -> cutp v p i (use h s) r2
    -> exists h' s',
         r2 = use h' s'
         /\ rsubs_head i (bound p) h h'
         /\ cutp v p i s s'.
Proof.
intros v p i h s r2 j k Hlower Hupper Hrigid Hcutp.
revert v i h s r2 j k Hlower Hupper Hrigid Hcutp.
induct p.

(* down *)
intros v i h s r2 j k Hlower Hupper Hrigid Hcutp.
simpl in Hupper.
invertc Hcutp; [].
intros m Hsub <-.
so (rigid_adjust _ _ i 1 _ Hlower Hupper Hrigid) as Hrigid'.
sor (refine (rsub_rigid_invert _#6 Hsub Hrigid' _ _)) as Hinst; [omega | omega |].
clear Hrigid'.
invertc Hinst; [].
intros h' s' Hsubh Hsubs <-.
exists h', s'.
simpl.
do2 2 split; eauto using @cutp; [].
#2 invert Hsubh; try (intros; subst; apply rsubs_const; done); invertc Hrigid; try (intros; subst; discriminate).
  intros l Hineq <- l' Hlt Heq <-.
  injection Heq; clear Heq; intros; subst l'.
  apply rsubs_var_lt; omega.

  intros l Hineq <- l' Hgt Heq <-.
  injection Heq; clear Heq; intros; subst l'.
  apply rsubs_var_gt; omega.

(* one *)
intros v i h s r2 j k Hlower Hupper Hrigid Hcutp.
invert Hcutp; [].
intros <- <-.
exists h, s.
simpl.
do2 2 split; eauto using @cutp; [].
invert Hrigid; try (intros; subst; eauto using rsubs_head; done); [].
intros l Hineq <-.
#2 destruct Hineq.
  apply rsubs_var_lt; omega.

  repl l with (l - 0) at 2 by omega.
  apply rsubs_var_gt; omega.

(* tens *)
intros p IH1 q IH2 vv i h s r2 j k Hlower Hupper Hrigid Hcutp.
simpl in Hupper.
invertc Hcutp; [].
intros v w r1 Hcut1 Hcut2 <-.
sor (refine (IH1 _#7 _ _ Hrigid Hcut1)) as (h1 & s1 & -> & Hsubh1 & Hcut1'); try omega; [].
sor (refine (IH2 _#7 _ _ (rsubs_head_rigid _#8 _ _ Hrigid Hsubh1) Hcut2)) as (h2 & s2 & -> & Hsubh2 & Hcut2'); try omega; [].
exists h2, s2.
do2 2 split; eauto using @cutp; [].
simpl.
invert Hsubh1; invert Hsubh2;
try (intros; subst;
     repeat
       (match goal with
        | Heq : _ = _ |- _ => injection Heq; clear Heq
        end);
     intros; subst; try discriminate; try omega; eauto using rsubs_head; done); [].
intros ? H2 <- <- l H0 <- Heq.
injection Heq; [].
intros <-.
repl (l - bound p - bound q) with (l - (bound p + bound q)) by omega.
apply rsubs_var_gt; omega.
Qed.


Lemma cutp_rigid :
  forall v p i r r' j k,
    j <= i
    -> i + bound p <= j + k
    -> rigid j k r
    -> cutp v p i r r'
    -> rigid j (k - bound p) r'.
Proof.
intros v p i r r' j k Hlower Hupper Hrigid Hcut.
assert (exists h s, r = use h s) as (h & s & ->).
  invert Hrigid; intros; subst; eauto; done.
so (cutp_rigid_omnibus _#8 Hlower Hupper Hrigid Hcut) as (h' & s' & -> & Hsubh & _).
eapply rsubs_head_rigid; eauto; omega.
Qed.


Lemma cutp_use_invert :
  forall v p i h s r,
    rigid i (bound p) (use h s)
    -> cutp v p i (use h s) r
    -> exists h' s',
         r = use h' s'
         /\ rsubs_head i (bound p) h h'
         /\ cutp v p i s s'.
Proof.
intros v p i h s r Hrigid Hcut.
sor (refine (cutp_rigid_omnibus _#6 i (bound p) _ _ Hrigid Hcut)) as H; try omega; [].
destruct H as (h' & s' & -> & Hsubh & Hsub).
eauto.
Qed.


Lemma cutp_nres :
  forall v p i r r',
    rigid i (bound p) r
    -> cutp v p i r r'
    -> nres r
    -> nres r'.
Proof.
intros v p i r r' Hrigid Hsub His.
invert His; [].
intros h s His' <-.
so (cutp_use_invert _#6 Hrigid Hsub) as (h' & s' & -> & _ & Hsubs).
apply nres_use; eauto using cutp_nspine; done.
Qed.


(* Subsituting through composition *)

Lemma sub_into_compose :
  forall f i s1 s2 s s1' s2',
    nspine s1
    -> compose s1 s2 s
    -> rsub f i s1 s1'
    -> rsub f i s2 s2'
    -> exists s', 
         rsub f i s s'
         /\ compose s1' s2' s'.
Proof.
intros f i s1 s2 s s1' s2' His Hcompose Hsubs1 Hsubs2.
revert f i s2 s s1' s2' Hcompose Hsubs1 Hsubs2.
induct His.

(* nl *)
intros f i s2 s s1' s2' Hcompose Hsubs1 Hsubs2.
invert Hsubs1; [].
intros <-.
invert Hcompose; [].
intros <-.
eexists; eauto using @compose; done.

(* app *)
intros v s1 _ IH f i s2 vs1 ss s2' Hcompose Hsubs1 Hsubs2.
invert Hsubs1; clear Hsubs1; [].
intros v' s1' Hsubv Hsubs1 <-.
invert Hcompose; clear Hcompose; [].
intros s Hcompose <-.
so (IH _#6 Hcompose Hsubs1 Hsubs2) as (s' & Hsubsts & Hcompose').
eexists; eauto using @rsub, @compose; done.

(* pi1 *)
intros s1 _ IH f i s2 ps ss s2' Hcompose Hsubs1 Hsubs2.
invert Hsubs1; clear Hsubs1; [].
intros s1' Hsubs1 <-.
invert Hcompose; clear Hcompose; [].
intros s Hcompose <-.
so (IH _#6 Hcompose Hsubs1 Hsubs2) as (s' & Hsubsts & Hcompose').
eexists; eauto using @rsub, @compose; done.

(* pi2 *)
intros s1 _ IH f i s2 ps ss s2' Hcompose Hsubs1 Hsubs2.
invert Hsubs1; clear Hsubs1; [].
intros s1' Hsubs1 <-.
invert Hcompose; clear Hcompose; [].
intros s Hcompose <-.
so (IH _#6 Hcompose Hsubs1 Hsubs2) as (s' & Hsubsts & Hcompose').
eexists; eauto using @rsub, @compose; done.
Qed.


Lemma inst_into_compose :
  forall f i r1 s2 r r1' s2',
    nres r1
    -> compose r1 s2 r
    -> inst f i r1 r1'
    -> rsub f i s2 s2'
    -> exists r',
         inst f i r r'
         /\ compose r1' s2' r'.
Proof.
intros f i r1 s2 r r1' s2' His Hcompose Hinst Hsubs2.
invertc His; [].
intros h s1 Hiss1 <-.
invertc Hinst; [].
intros h' s1' Hsubh Hsubs1 <-.
invertc Hcompose; [].
intros s Hcompose <-.
so (sub_into_compose _#7 Hiss1 Hcompose Hsubs1 Hsubs2) as (s' & Hsubs & Hcompose').
eexists; eauto using inst, @compose.
Qed.


Lemma cutp_into_rcompose :
  forall v p i r1 s2 r r1' s2',
    nres r1
    -> compose r1 s2 r
    -> rigid i (bound p) r1
    -> cutp v p i r1 r1'
    -> cutp v p i s2 s2'
    -> exists r',
         cutp v p i r r'
         /\ compose r1' s2' r'.
Proof.
intros v p i r1 s2 r r1' s2' Hisr1 Hcompose Hrigid Hsubr1 Hsubs2.
revert v i r1 s2 r r1' s2' Hisr1 Hcompose Hrigid Hsubr1 Hsubs2.
induct p.

(* down *)
intros v i r1 s2 r r1' s2' Hisr1 Hcompose Hrigid Hsubr1 Hsubs2.
simpl in Hrigid.
invertc Hsubr1; [].
intros m Hsubr1 <-.
invertc Hsubs2; [].
intros Hsubs2.
sor (refine (rsub_rigid_invert _#6 Hsubr1 Hrigid _ _)) as Hinst; try omega; [].
so (inst_into_compose _#7 Hisr1 Hcompose Hinst Hsubs2) as (r' & Hinst' & Hcompose').
exists r'; split; eauto using @cutp, inst_impl_rsub; done.

(* one *)
intros v i r1 s2 r r1' s2' Hisr1 Hcompose Hrigid Hsubr1 Hsubs2.
invertc Hsubr1; [].
intros <- <-.
invertc Hsubs2; [].
intros <-.
exists r; split; eauto using @cutp; done.

(* tens *)
intros p IH1 q IH2 vv i r1 s2 r r1' s2' Hisr1 Hcompose Hrigid Hsubr1 Hsubs2.
simpl in Hrigid.
invertc Hsubr1; [].
intros v w r1a Hsubr1a Hsubr1b <-.
invertc Hsubs2; [].
intros s2a Hsubs2a Hsubs2b.
assert (rigid (bound q + i) (bound p) r1) as Hrigid'.
  refine (rigid_adjust _#5 _ _ Hrigid); omega.
so (IH1 _#7 Hisr1 Hcompose Hrigid' Hsubr1a Hsubs2a) as (ra & Hsubra & Hcompose').
so (cutp_nres _#5 Hrigid' Hsubr1a Hisr1) as Hisr1a.
sor (refine (cutp_rigid _#7 _ _ Hrigid Hsubr1a)) as Hrigid1a; try omega; [].
repl (bound p + bound q - bound p) with (bound q) in Hrigid1a by omega.
so (IH2 _#7 Hisr1a Hcompose' Hrigid1a Hsubr1b Hsubs2b) as (rb & Hsubrb & Hcompose'').
exists rb; split; eauto using @cutp; done.
Qed.


Lemma sub_into_compose_reverse :
  forall f i s1 s2 s s',
    nspine s1
    -> compose s1 s2 s
    -> rsub f i s s'
    -> (exists s1', rsub f i s1 s1')
       /\
       (exists s2', rsub f i s2 s2').
Proof.
intros f i s1 s2 s s' His Hcompose Hsub.
revert f i s2 s s' Hcompose Hsub.
induct His.

(* nl *)
intros f i s2 s s' Hcompose Hsub.
invert Hcompose; [].
intros <-.
eauto using @rsub; done.

(* app *)
intros v s1 _ IH f i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
invert Hsub; clear Hsub; [].
intros v' s3' Hsubv Hsubs3 <-.
so (IH _#5 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
eauto using @rsub; done.

(* pi1 *)
intros s1 _ IH f i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
invert Hsub; clear Hsub; [].
intros s3' Hsubs3 <-.
so (IH _#5 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
eauto using @rsub; done.

(* pi2 *)
intros s1 _ IH f i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
invert Hsub; clear Hsub; [].
intros s3' Hsubs3 <-.
so (IH _#5 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
eauto using @rsub; done.
Qed.


Lemma sub_into_rcompose_reverse :
  forall f i r1 s2 r r',
    nres r1
    -> compose r1 s2 r
    -> rsub f i r r'
    -> exists s2', rsub f i s2 s2'.
Proof.
intros f i r1 s2 r r' His Hcompose Hsub.
invert His; clear His; [].
intros h s1 His' <-.
invert Hcompose; clear Hcompose; [].
intros s Hcompose' <-.
assert (exists s', rsub f i s s') as (s' & Hsubs).
  invert Hsub; intros; subst; eexists; eauto; done.
eapply sub_into_compose_reverse; eauto; done.
Qed.


Lemma cutp_into_compose_reverse :
  forall v p i s1 s2 s s',
    nspine s1
    -> compose s1 s2 s
    -> cutp v p i s s'
    -> (exists s1', cutp v p i s1 s1')
       /\
       (exists s2', cutp v p i s2 s2').
Proof.
intros v p i s1 s2 s s' His Hcompose Hsub.
revert v p i s2 s s' Hcompose Hsub.
induct His.

(* nl *)
intros v p i s2 s s' Hcompose Hsub.
invert Hcompose; [].
intros <-.
split; eauto using @rsub; [].
exists nl.
eapply cutp_compat0; eauto using cutp_patmatch, @rsub; done.

(* app *)
intros v s1 _ IH w p i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
#2 exploit (cutp_invert2 app) as Hinvert.
  intros f i0 d1 d2 e H; invert H; intros; eauto; done.
so (Hinvert _#6 Hsub) as (v' & s3' & -> & Hsubv & Hsubs3).
so (IH _#6 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
split; eauto using @rsub, cutp_compat2; done.

(* pi1 *)
intros s1 _ IH w p i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
#2 exploit (cutp_invert1 pi1) as Hinvert.
  intros f i0 d e H; invert H; intros; eauto; done.
so (Hinvert _#5 Hsub) as (s3' & -> & Hsubs3).
so (IH _#6 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
split; eauto using @rsub, cutp_compat1; done.

(* pi2 *)
intros s1 _ IH w p i s2 s s' Hcompose Hsub.
invert Hcompose; clear Hcompose; [].
intros s3 Hcompose' <-.
#2 exploit (cutp_invert1 pi2) as Hinvert.
  intros f i0 d e H; invert H; intros; eauto; done.
so (Hinvert _#5 Hsub) as (s3' & -> & Hsubs3).
so (IH _#6 Hcompose' Hsubs3) as ((s1', Hsubs1), (s2', Hsubs2)).
split; eauto using @rsub, cutp_compat1; done.
Qed.


Lemma cutp_into_rcompose_reverse :
  forall v p i r1 s2 r r',
    nres r1
    -> rigid i (bound p) r1
    -> compose r1 s2 r
    -> cutp v p i r r'
    -> exists s2', cutp v p i s2 s2'.
Proof.
intros v p i r1 s2 r r' His Hrigid Hcompose Hsub.
invert His; clear His; [].
intros h s1 His' <-.
invert Hcompose; clear Hcompose; [].
intros s Hcompose' <-.
so (cutp_use_invert _#6 (rigid_compose _#5 (compose_use Hcompose') Hrigid) Hsub) as (_ & s' & _ & _ & Hsubs).
eapply cutp_into_compose_reverse; eauto.
Qed.
